import datetime
from dataclasses import dataclass
from typing import Any, List, Optional, Set, Union

import numpy as np

from valor.schemas.geometry import (
    BoundingBox,
    MultiPolygon,
    Point,
    Polygon,
    Raster,
)
from valor.typing import is_geojson

GeometryType = Union[
    Point,
    Polygon,
    BoundingBox,
    MultiPolygon,
    Raster,
]


@dataclass
class Constraint:
    """
    Represents a constraint with a value and an operator.

    Attributes:
        value : Any
            The value associated with the constraint.
        operator : str
            The operator used to define the constraint.
    """

    value: Any
    operator: str


@dataclass
class BinaryExpression:
    """
    Stores a conditional relationship.

    Attributes
    ----------
    name : str
        The name of the filter property.
    constraint : valor.schemas.Constraint
        The operation that is performed.
    key : str, optional
        An optional key used for object retrieval.
    """

    name: str
    constraint: Constraint
    key: Union[str, None] = None


class _DeclarativeMapper:
    """
    Base class for constructing mapping objects.

    Parameters
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def __init__(
        self,
        name: str,
        key: Optional[str] = None,
    ):
        self.name = name
        self.key = key
        self.valid_operators = {}

    def _valid_operators(self) -> Set[str]:
        """
        Set of valid operators.

        Overload in subclasses with the following:
        >>> valid_operators = {op1, op2}
        >>> return super()._valid_operators().union(valid_operators)
        """
        return set()

    def _create_expression(
        self, value: Any, operator: str
    ) -> BinaryExpression:
        """Called by operator overload functions."""
        if operator not in self._valid_operators():
            raise AttributeError(
                f"Mapper with type `{type(self)}` does not suppoert the `{operator}` operator."
            )

        # overload-able validation
        self._validate(value=value, operator=operator)

        # overload-able modifications
        name = self._modify_name(operator=operator)
        value = self._modify_value(value=value, operator=operator)

        return BinaryExpression(
            name=name,
            key=self.key,
            constraint=Constraint(
                value=value,
                operator=operator,
            ),
        )

    def _validate(self, value: Any, operator: str) -> None:
        """Overload in subclasses to insert a validator."""
        pass

    def _modify_value(self, value: Any, operator: str) -> Any:
        """Overload in subclasses to insert a value modification."""
        return value

    def _modify_name(self, operator: str) -> str:
        """Overload in subclasses to insert a name modification."""
        return self.name

    def __eq__(self, value: Any) -> BinaryExpression:  # type: ignore
        raise AttributeError(f"'{type(self)}' object has no attribute '=='")

    def __ne__(self, value: Any) -> BinaryExpression:  # type: ignore
        raise AttributeError(f"'{type(self)}' object has no attribute '!='")

    def __lt__(self, value: Any) -> BinaryExpression:
        raise AttributeError(f"'{type(self)}' object has no attribute '<'")

    def __gt__(self, value: Any) -> BinaryExpression:
        raise AttributeError(f"'{type(self)}' object has no attribute '>'")

    def __le__(self, value: Any) -> BinaryExpression:
        raise AttributeError(f"'{type(self)}' object has no attribute '<='")

    def __ge__(self, value: Any) -> BinaryExpression:
        raise AttributeError(f"'{type(self)}' object has no attribute '>='")


class _NullableMapper(_DeclarativeMapper):
    """
    Defines a mapping object that handles nullable values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _valid_operators(self) -> Set[str]:
        valid_operators = {"is_none", "exists"}
        return super()._valid_operators().union(valid_operators)

    def is_none(self) -> BinaryExpression:
        return self._create_expression(None, "is_none")

    def exists(self) -> BinaryExpression:
        return self._create_expression(None, operator="exists")


class _EquatableMapper(_DeclarativeMapper):
    """
    Defines a mapping object that handles equatable values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _valid_operators(self) -> Set[str]:
        valid_operators = {"==", "!="}
        return super()._valid_operators().union(valid_operators)

    def __eq__(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "==")

    def __ne__(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "!=")

    def in_(self, values: List[Any]) -> List[BinaryExpression]:
        if not isinstance(values, list):
            raise TypeError("`in_` takes a list as input.")
        return [self == value for value in values]


class _QuantifiableMapper(_EquatableMapper):
    """
    Defines a mapping object that handles quantifiable values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _valid_operators(self) -> Set[str]:
        valid_operators = {">", "<", ">=", "<=", "==", "!="}
        return super()._valid_operators().union(valid_operators)

    def __lt__(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "<")

    def __gt__(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, ">")

    def __le__(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "<=")

    def __ge__(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, ">=")


class BoolMapper(_EquatableMapper):
    """
    Defines a mapping object that handles boolean values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _validate(self, value: Any, operator: str) -> None:
        if not isinstance(value, bool):
            raise TypeError(
                f"BoolMapper does not support object type `{type(value)}`."
            )


class StringMapper(_EquatableMapper):
    """
    Defines a mapping object that handles string values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _validate(self, value: Any, operator: str) -> None:
        if not isinstance(value, str):
            raise TypeError(
                f"StringMapper does not support object type `{type(value)}`."
            )


class NumericMapper(_QuantifiableMapper):
    """
    Defines a mapping object that handles numeric values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _validate(self, value: Any, operator: str) -> None:
        if type(value) not in [int, float, np.floating]:
            raise TypeError(
                f"NumericMapper does not support object type `{type(value)}`."
            )


class DatetimeMapper(_QuantifiableMapper):
    """
    Defines a mapping object that handles `datetime` objects.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _modify_value(self, value: Any, operator: str) -> Any:
        vtype = type(value)
        if vtype is datetime.datetime:
            return {"datetime": value.isoformat()}
        elif vtype is datetime.date:
            return {"date": value.isoformat()}
        elif vtype is datetime.time:
            return {"time": value.isoformat()}
        elif vtype is datetime.timedelta:
            return {"duration": str(value.total_seconds())}
        else:
            raise TypeError(
                f"DatetimeMapper does not support object type `{type(value)}`."
            )


class _SpatialMapper(_NullableMapper):
    """
    Defines a mapping object that handles spatial values.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    @property
    def area(self):
        return NumericMapper(name=f"{self.name}_area")

    def _valid_operators(self) -> Set[str]:
        valid_operators = {"contains", "inside", "outside", "intersect"}
        return super()._valid_operators().union(valid_operators)

    def contains(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "contains")

    def intersect(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "intersect")

    def inside(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "inside")

    def outside(self, value: Any) -> BinaryExpression:
        return self._create_expression(value, "outside")


class GeometryMapper(_SpatialMapper):
    """
    Defines a mapping object that handles `valor.schemas.geometry` objects.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _validate(self, value: GeometryType, operator: str):

        if operator in {"is_none", "exists"}:
            return

        if type(value) not in [
            Point,
            BoundingBox,
            Polygon,
            MultiPolygon,
            Raster,
        ]:
            raise TypeError(
                f"GeometryMapper does not support objects of type `{type(value)}`."
            )

        raise NotImplementedError(
            "Geometric types only support 'is_none' and 'exists'. Support for other spatial operators is planned."
        )

    def _modify_name(self, operator: str) -> str:
        """Rename geometric type when nullable operators are used."""
        if operator in {"is_none", "exists"}:
            return f"require_{self.name}"
        return self.name


class GeospatialMapper(_SpatialMapper):
    """
    Defines a mapping object that handles GeoJSON.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _validate(self, value: GeometryType, operator: str):

        if operator not in {"inside", "outside", "intersect"}:
            raise NotImplementedError(
                "Geospatial types only support 'inside', 'outside' and 'intersect'. Support for other spatial operators is planned."
            )

        if not isinstance(value, dict):
            raise TypeError(
                "Geospatial values should be a GeoJSON-style dictionary containing the keys `type` and `coordinates`."
            )
        elif not value.get("type") or not value.get("coordinates"):
            raise KeyError(
                "Geospatial values should be a GeoJSON-style dictionary containing the keys `type` and `coordinates`."
            )


class _DictionaryValueMapper(
    _QuantifiableMapper, _SpatialMapper, _NullableMapper
):
    """
    Defines a mapping object that handles arbitrary objects.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _create_expression(self, value: Any, operator: str) -> Any:
        if self.key is None:
            raise ValueError(
                "Attribute `key` is required for `_DictionaryValueMapper`."
            )

        if operator in {"is_none", "exists"} and value is None:
            return BinaryExpression(
                name=self.name,
                key=self.key,
                constraint=Constraint(
                    value=None,
                    operator=operator,
                ),
            )

        # direct value to appropriate mapper (if it exists)
        vtype = type(value)
        if vtype is bool:
            return BoolMapper(name=self.name, key=self.key)._create_expression(
                value, operator
            )
        if vtype is str:
            return StringMapper(
                name=self.name, key=self.key
            )._create_expression(value, operator)
        elif vtype in [int, float]:
            return NumericMapper(
                name=self.name, key=self.key
            )._create_expression(value, operator)
        elif vtype in [
            datetime.datetime,
            datetime.date,
            datetime.time,
            datetime.timedelta,
        ]:
            return DatetimeMapper(
                name=self.name, key=self.key
            )._create_expression(value, operator)
        elif is_geojson(value):
            return GeospatialMapper(
                name=self.name, key=self.key
            )._create_expression(value, operator)
        else:
            raise NotImplementedError(
                f"DictionaryMapper value `{value}` is not supported."
            )


class DictionaryMapper(_DeclarativeMapper):
    """
    Defines a dictionary mapping object that captures the key used to access the data..

    This object returns a `_DictionaryValueMapper` when given a key.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.

    Examples
    --------
    >>> DictionaryMapper("name")["some_key"] == True
    BinaryExpression(name='name', constraint=Constraint(value=True, operator='=='), key='some_key')
    """

    def __getitem__(self, key: str):
        return _DictionaryValueMapper(name=self.name, key=key)

    def _create_expression(
        self, value: Any, operator: str
    ) -> BinaryExpression:
        raise NotImplementedError(
            "Dictionary mapper does not define any operations for iteself. Please use `dict[key]` to create an expression."
        )


class LabelMapper(_EquatableMapper):
    """
    Defines a mapping object that handles `valor.Label` objects.

    This object maps conditional expressions into `BinaryExpression` objects.

    Attributes
    ----------
    name : str
        The name of the filter property.
    key : str, optional
        An optional key used for object retrieval.
    """

    def _modify_value(self, value: Any, operator: str) -> Any:
        # convert to dict
        if hasattr(value, "to_dict"):
            value = value.to_dict()

        # validate dict
        if not isinstance(value, dict):
            raise TypeError(
                "Label must be a `dict` or `valor.Label` that contains `key` and `value` attributes."
            )
        elif not set(value.keys()).issuperset({"key", "value"}):
            raise KeyError("Label must contain `key` and `value` keys.")
        elif type(value["key"]) is not str:
            raise ValueError("Label key must be of type `str`.")
        elif type(value["value"]) is not str:
            raise ValueError("Label value must be of type `str`.")

        return {value["key"]: value["value"]}



import sys

if sys.version_info.minor >= 8:
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata  # type: ignore

try:
    __version__ = importlib_metadata.version("valor")
except importlib_metadata.PackageNotFoundError:
    __version__ = ""

from .client import ClientConnection, connect
from .coretypes import (
    Annotation,
    Client,
    Dataset,
    Datum,
    Evaluation,
    GroundTruth,
    Label,
    Model,
    Prediction,
)
from .schemas import Constraint, Filter

__all__ = [
    "connect",
    "Client",
    "ClientConnection",
    "Label",
    "Evaluation",
    "Dataset",
    "Model",
    "Datum",
    "Annotation",
    "GroundTruth",
    "Prediction",
    "Filter",
    "Constraint",
]



from __future__ import annotations

import json
import time
import warnings
from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from valor.client import ClientConnection, connect, get_connection
from valor.enums import AnnotationType, EvaluationStatus, TableStatus, TaskType
from valor.exceptions import ClientException
from valor.schemas.constraints import BinaryExpression
from valor.schemas.evaluation import EvaluationParameters, EvaluationRequest
from valor.schemas.filters import Filter
from valor.schemas.geometry import BoundingBox, Polygon, Raster
from valor.schemas.metadata import (
    MetadataType,
    dump_metadata,
    load_metadata,
    validate_metadata,
)
from valor.schemas.properties import (
    DictionaryProperty,
    GeometryProperty,
    LabelProperty,
    NumericProperty,
    StringProperty,
)
from valor.typing import is_float

FilterType = Union[
    Filter, List[Union[BinaryExpression, List[BinaryExpression]]], dict
]


def _format_filter(filter_by: Optional[FilterType]) -> Filter:
    """
    Formats the various filter or constraint representations into a 'schemas.Filter' object.

    Parameters
    ----------
    filter_by : FilterType, optional
        The reference filter.

    Returns
    -------
    valor.schemas.Filter
        A properly formatted 'schemas.Filter' object.
    """
    if isinstance(filter_by, Filter):
        return filter_by
    elif isinstance(filter_by, list) or filter_by is None:
        filter_by = filter_by if filter_by else []
        return Filter.create(filter_by)
    elif isinstance(filter_by, dict):
        return Filter(**filter_by)
    else:
        raise TypeError


class Label:
    """
    An object for labeling datasets, models, and annotations.

    Parameters
    ----------
    key : str
        The class key of the label.
    value : str
        The class value of the label.
    score : float, optional
        The score associated with the label (if applicable).

    Attributes
    ----------
    filter_by : filter_factory
        Declarative mappers used to create filters.
    """

    value = StringProperty("label_values")
    key = StringProperty("label_keys")
    score = NumericProperty("label_scores")

    def __init__(
        self,
        key: str,
        value: str,
        score: Union[float, np.floating, None] = None,
    ):
        if not isinstance(key, str):
            raise TypeError("Attribute `key` should have type `str`.")
        if not isinstance(value, str):
            raise TypeError("Attribute `value` should have type `str`.")
        if score is not None:
            if not is_float(score):
                raise TypeError(
                    "Attribute `score` should be a floating-point number or `None`."
                )

        self.key = key
        self.value = value
        self.score = score

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        return json.dumps(self.to_dict(), indent=4)

    def __repr__(self) -> str:
        return str(self.tuple())

    def to_dict(self) -> Dict[str, Union[str, float, np.floating, None]]:
        """
        Defines how a `valor.Label` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing a label.
        """
        return {
            "key": self.key,
            "value": self.value,
            "score": self.score,
        }

    @classmethod
    def from_dict(cls, resp) -> Label:
        """
        Construct a label from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing a label.

        Returns
        -------
        valor.Label
        """
        return cls(**resp)

    def __eq__(self, other) -> bool:
        """
        Defines how `Labels` are compared to one another.

        Parameters
        ----------
        other : Label
            The object to compare with the `Label`.

        Returns
        ----------
        boolean
            A boolean describing whether the two objects are equal.
        """
        # type mismatch
        if type(other) is not type(self):
            return False

        # k,v mismatch
        if self.key != other.key or self.value != other.value:
            return False

        # score is None
        if self.score is None or other.score is None:
            return (other.score is None) == (self.score is None)

        # scores not equal
        if is_float(self.score) and is_float(other.score):
            return bool(np.isclose(self.score, other.score))

        return False

    def __hash__(self) -> int:
        """
        Defines how a `Label` is hashed.

        Returns
        ----------
        int
            The hashed 'Label`.
        """
        return hash(f"key:{self.key},value:{self.value},score:{self.score}")

    def tuple(self) -> Tuple[str, str, Union[float, np.floating, None]]:
        """
        Defines how the `Label` is turned into a tuple.

        Returns
        ----------
        tuple
            A tuple of the `Label's` arguments.
        """
        return (self.key, self.value, self.score)


class Annotation:
    """
    A class used to annotate `GroundTruths` and `Predictions`.

    Parameters
    ----------
    task_type: TaskType
        The task type associated with the `Annotation`.
    labels: List[Label], optional
        A list of labels to use for the `Annotation`.
    metadata: Dict[str, Union[int, float, str, bool, datetime.datetime, datetime.date, datetime.time]]
        A dictionary of metadata that describes the `Annotation`.
    bounding_box: BoundingBox, optional
        A bounding box to assign to the `Annotation`.
    polygon: Polygon, optional
        A polygon to assign to the `Annotation`.
    raster: Raster, optional
        A raster to assign to the `Annotation`.
    embedding: List[float], optional
        An embedding, described by a list of values with type float and a maximum length of 16,000.

    Attributes
    ----------
    geometric_area : float
        The area of the annotation.

    Examples
    --------

    Classification
    >>> Annotation(
    ...     task_type=TaskType.CLASSIFICATION,
    ...     labels=[
    ...         Label(key="class", value="dog"),
    ...         Label(key="category", value="animal"),
    ...     ]
    ... )

    Object-Detection BoundingBox
    >>> annotation = Annotation(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     bounding_box=box2,
    ... )

    Object-Detection Polygon
    >>> annotation = Annotation(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     polygon=polygon1,
    ... )

    Object-Detection Raster
    >>> annotation = Annotation(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     raster=raster1,
    ... )

    Semantic-Segmentation Raster
    >>> annotation = Annotation(
    ...     task_type=TaskType.SEMANTIC_SEGMENTATION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     raster=raster1,
    ... )

    Defining all supported annotation types for a given `task_type` is allowed!
    >>> Annotation(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     bounding_box=box1,
    ...     polygon=polygon1,
    ...     raster=raster1,
    ... )
    """

    task_type = StringProperty("task_types")
    labels = LabelProperty("labels")
    metadata = DictionaryProperty("annotation_metadata")
    bounding_box = GeometryProperty("bounding_box")
    polygon = GeometryProperty("polygon")
    raster = GeometryProperty("raster")

    def __init__(
        self,
        task_type: TaskType,
        labels: Optional[List[Label]] = None,
        metadata: Optional[MetadataType] = None,
        bounding_box: Optional[BoundingBox] = None,
        polygon: Optional[Polygon] = None,
        raster: Optional[Raster] = None,
        embedding: Optional[List[float]] = None,
    ):
        self.task_type = TaskType(task_type)
        self.labels = labels if labels else []
        self.metadata = metadata if metadata else {}
        self.bounding_box = bounding_box
        self.polygon = polygon
        self.raster = raster
        self.embedding = embedding
        self._validate()

    def _validate(self):
        """
        Validates the parameters used to create an `Annotation` object.
        """
        # labels
        if not isinstance(self.labels, list):
            raise TypeError(
                "Attribute `labels` should have type `List[valor.Label]`."
            )
        for idx, label in enumerate(self.labels):
            if not isinstance(label, Label):
                raise TypeError(
                    f"Attribute `labels[{idx}]` should have type `valor.Label`."
                )

        # bounding box
        if self.bounding_box is not None:
            if not isinstance(self.bounding_box, BoundingBox):
                raise TypeError(
                    "Attribute `bounding_box` should have type `valor.schemas.BoundingBox`."
                )

        # polygon
        if self.polygon is not None:
            if not isinstance(self.polygon, Polygon):
                raise TypeError(
                    "Attribute `polygon` should have type `valor.schemas.Polygon`."
                )

        # raster
        if self.raster is not None:
            if not isinstance(self.raster, Raster):
                raise TypeError(
                    "Attribute `raster` should have type `valor.schemas.Raster`."
                )

        # embedding
        if self.embedding is not None:
            if not isinstance(self.embedding, list):
                raise TypeError(
                    "Attribute `embedding` should have type `List[float]`."
                )
            for idx, embedding in enumerate(self.embedding):
                if not isinstance(embedding, (int, float, np.floating)):
                    raise TypeError(
                        f"Attribute `embedding[{idx}]` should have type `float`, received element with type `{type(embedding)}`."
                    )

        # metadata
        if not isinstance(self.metadata, dict):
            raise TypeError("Attribute `metadata` should have type `dict`.")
        validate_metadata(self.metadata)

    @classmethod
    def from_dict(cls, resp: dict) -> Annotation:
        """
        Construct an annotation from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing an annotation.

        Returns
        -------
        valor.Annotation
        """

        task_type = TaskType(resp["task_type"])
        labels = [Label.from_dict(label) for label in resp["labels"]]
        metadata = load_metadata(resp["metadata"])

        bounding_box = None
        polygon = None
        raster = None
        embedding = None

        if "bounding_box" in resp:
            bounding_box = (
                BoundingBox(**resp["bounding_box"])
                if resp["bounding_box"]
                else None
            )
        if "polygon" in resp:
            polygon = Polygon(**resp["polygon"]) if resp["polygon"] else None
        if "raster" in resp:
            raster = Raster(**resp["raster"]) if resp["raster"] else None
        if "embedding" in resp:
            embedding = resp["embedding"]

        return cls(
            task_type=task_type,
            labels=labels,
            metadata=metadata,
            bounding_box=bounding_box,
            polygon=polygon,
            raster=raster,
            embedding=embedding,
        )

    def to_dict(self) -> dict:
        """
        Defines how a `valor.Annotation` object is serialized into a dictionary.

        Returns
        -------
        dict
            A dictionary describing an annotation.
        """
        return {
            "task_type": self.task_type.value,
            "labels": [label.to_dict() for label in self.labels],
            "metadata": dump_metadata(self.metadata),
            "bounding_box": (
                asdict(self.bounding_box) if self.bounding_box else None
            ),
            "polygon": asdict(self.polygon) if self.polygon else None,
            "raster": asdict(self.raster) if self.raster else None,
            "embedding": self.embedding if self.embedding else None,
        }

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        return json.dumps(self.to_dict(), indent=4)

    def __eq__(self, other) -> bool:
        """
        Defines how `Annotations` are compared to one another.

        Parameters
        ----------
        other : Annotation
            The object to compare with the `Annotation`.

        Returns
        ----------
        boolean
            A boolean describing whether the two objects are equal.
        """
        if not isinstance(other, Annotation):
            raise TypeError(
                f"Expected type `{type(Annotation)}`, got `{other}`"
            )
        return self.to_dict() == other.to_dict()


class Datum:
    """
    A class used to store datum about `GroundTruths` and `Predictions`.

    Parameters
    ----------
    uid : str
        The UID of the `Datum`.
    metadata : dict
        A dictionary of metadata that describes the `Datum`.
    """

    uid = StringProperty("datum_uids")
    metadata = DictionaryProperty("datum_metadata")

    def __init__(
        self,
        uid: str,
        metadata: Optional[MetadataType] = None,
    ):
        self.uid = uid
        self.metadata = metadata if metadata else {}

        if not isinstance(self.uid, str):
            raise TypeError("Attribute `uid` should have type `str`.")
        validate_metadata(self.metadata)

    def to_dict(self, dataset_name: Optional[str] = None) -> dict:
        """
        Defines how a `valor.Datum` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing a datum.
        """
        return {
            "dataset_name": dataset_name,
            "uid": self.uid,
            "metadata": dump_metadata(self.metadata),
        }

    @classmethod
    def from_dict(cls, resp: dict) -> Datum:
        """
        Construct a datum from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing a datum.

        Returns
        -------
        valor.Datum
        """
        resp.pop("dataset_name", None)
        resp["metadata"] = load_metadata(resp["metadata"])
        return cls(**resp)

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        objdict = self.to_dict(dataset_name=None)
        objdict.pop("dataset_name")
        return json.dumps(objdict, indent=4)

    def __eq__(self, other) -> bool:
        """
        Defines how `Datums` are compared to one another.

        Parameters
        ----------
        other : Datum
            The object to compare with the `Datum`.

        Returns
        ----------
        bool
            A boolean describing whether the two objects are equal.
        """
        if not isinstance(other, Datum):
            raise TypeError(f"Expected type `{type(Datum)}`, got `{other}`")
        return self.to_dict() == other.to_dict()


class GroundTruth:
    """
    An object describing a ground truth (e.g., a human-drawn bounding box on an image).

    Parameters
    ----------
    datum : Datum
        The `Datum` associated with the `GroundTruth`.
    annotations : List[Annotation]
        The list of `Annotations` associated with the `GroundTruth`.
    """

    def __init__(self, datum: Datum, annotations: List[Annotation]):
        self.datum = datum
        self.annotations = annotations
        self._validate()

    def _validate(self):
        """
        Validate the inputs of the `GroundTruth`.
        """
        # validate datum
        if not isinstance(self.datum, Datum):
            raise TypeError(
                "Attribute `datum` should have type `valor.Datum`."
            )

        # validate annotations
        if not isinstance(self.annotations, list):
            raise TypeError(
                "Attribute `datum` should have type `List[valor.Annotation]`."
            )
        for idx, annotation in enumerate(self.annotations):
            if not isinstance(annotation, Annotation):
                raise TypeError(
                    f"Attribute `annotations[{idx}]` should have type `valor.Annotation`."
                )

    def to_dict(
        self,
        dataset_name: Optional[str] = None,
    ) -> dict:
        """
        Defines how a `valor.GroundTruth` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing a ground truth.
        """
        return {
            "datum": self.datum.to_dict(dataset_name),
            "annotations": [
                annotation.to_dict() for annotation in self.annotations
            ],
        }

    @classmethod
    def from_dict(cls, resp: dict) -> GroundTruth:
        """
        Construct a ground truth from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing a ground truth.

        Returns
        -------
        valor.GroundTruth
        """
        expected_keys = {"datum", "annotations"}
        if set(resp.keys()) != expected_keys:
            raise ValueError(
                f"Expected keys `{expected_keys}`, received `{set(resp.keys())}`."
            )
        if not isinstance(resp["annotations"], list):
            raise TypeError("Expected `annotations` member to be a `list`.")
        return cls(
            datum=Datum.from_dict(resp["datum"]),
            annotations=[
                Annotation.from_dict(annotation)
                for annotation in resp["annotations"]
            ],
        )

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        objdict = self.to_dict(dataset_name=None)
        objdict["datum"].pop("dataset_name")
        return json.dumps(objdict, indent=4)

    def __eq__(self, other):
        """
        Defines how `GroundTruths` are compared to one another.

        Parameters
        ----------
        other : GroundTruth
            The object to compare with the `GroundTruth`.

        Returns
        ----------
        boolean
            A boolean describing whether the two objects are equal.
        """
        if not isinstance(other, GroundTruth):
            raise TypeError(
                f"Expected type `{type(GroundTruth)}`, got `{other}`"
            )
        return self.to_dict() == other.to_dict()


class Prediction:
    """
    An object describing a prediction (e.g., a machine-drawn bounding box on an image).

    Parameters
    ----------
    datum : Datum
        The `Datum` associated with the `Prediction`.
    annotations : List[Annotation]
        The list of `Annotations` associated with the `Prediction`.

    Attributes
    ----------
    score : Union[float, int]
        The score assigned to the `Prediction`.
    """

    def __init__(
        self,
        datum: Datum,
        annotations: List[Annotation],
    ):
        self.datum = datum
        self.annotations = annotations
        self._validate()

    def _validate(self):
        """
        Validate the inputs of the `Prediction`.
        """
        # validate datum
        if not isinstance(self.datum, Datum):
            raise TypeError(
                "Attribute `datum` should have type `valor.Datum`."
            )

        # validate annotations
        if not isinstance(self.annotations, list):
            raise TypeError(
                "Attribute `datum` should have type `List[valor.Annotation]`."
            )
        for idx, annotation in enumerate(self.annotations):
            if not isinstance(annotation, Annotation):
                raise TypeError(
                    f"Attribute `annotations[{idx}]` should have type `valor.Annotation`."
                )

        # TaskType-specific validations
        for annotation in self.annotations:
            if annotation.task_type in [
                TaskType.CLASSIFICATION,
                TaskType.OBJECT_DETECTION,
            ]:
                for label in annotation.labels:
                    if label.score is None:
                        raise ValueError(
                            f"For task type `{annotation.task_type}` prediction labels must have scores, but got `None`"
                        )
            if annotation.task_type == TaskType.CLASSIFICATION:
                label_keys_to_sum = {}
                for scored_label in annotation.labels:
                    label_key = scored_label.key
                    if label_key not in label_keys_to_sum:
                        label_keys_to_sum[label_key] = 0.0
                    label_keys_to_sum[label_key] += scored_label.score

                for k, total_score in label_keys_to_sum.items():
                    if abs(total_score - 1) > 1e-5:
                        raise ValueError(
                            "For each label key, prediction scores must sum to 1, but"
                            f" for label key {k} got scores summing to {total_score}."
                        )

    def to_dict(
        self,
        dataset_name: Optional[str] = None,
        model_name: Optional[str] = None,
    ) -> dict:
        """
        Defines how a `valor.Prediction` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing a prediction.
        """
        return {
            "datum": self.datum.to_dict(dataset_name=dataset_name),
            "model_name": model_name,
            "annotations": [
                annotation.to_dict() for annotation in self.annotations
            ],
        }

    @classmethod
    def from_dict(cls, resp: dict) -> Prediction:
        """
        Construct a Prediction from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing a prediction.

        Returns
        -------
        valor.Prediction
        """
        expected_keys = {"datum", "annotations", "model_name"}
        if set(resp.keys()) != expected_keys:
            raise ValueError(
                f"Expected keys `{expected_keys}`, received `{set(resp.keys())}`."
            )
        if not isinstance(resp["annotations"], list):
            raise TypeError("Expected `annotations` member to be a `list`.")
        return cls(
            datum=Datum.from_dict(resp["datum"]),
            annotations=[
                Annotation.from_dict(annotation)
                for annotation in resp["annotations"]
            ],
        )

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        objdict = self.to_dict(dataset_name=None)
        objdict["datum"].pop("dataset_name")
        objdict.pop("model_name")
        return json.dumps(objdict, indent=4)

    def __eq__(self, other):
        """
        Defines how `Predictions` are compared to one another.

        Parameters
        ----------
        other : Prediction
            The object to compare with the `Prediction`.

        Returns
        ----------
        boolean
            A boolean describing whether the two objects are equal.
        """
        if not isinstance(other, Prediction):
            raise TypeError(
                f"Expected type `{type(Prediction)}`, got `{other}`"
            )
        return self.to_dict(None, None) == other.to_dict(None, None)


class Evaluation:
    """
    Wraps `valor.client.Job` to provide evaluation-specifc members.
    """

    def __init__(
        self, connection: Optional[ClientConnection] = None, **kwargs
    ):
        """
        Defines important attributes of the API's `EvaluationResult`.

        Attributes
        ----------
        id : int
            The ID of the evaluation.
        model_name : str
            The name of the evaluated model.
        datum_filter : schemas.Filter
            The filter used to select the datums for evaluation.
        status : EvaluationStatus
            The status of the evaluation.
        metrics : List[dict]
            A list of metric dictionaries returned by the job.
        confusion_matrices : List[dict]
            A list of confusion matrix dictionaries returned by the job.
        """
        if not connection:
            connection = get_connection()
        self.conn = connection
        self.update(**kwargs)

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        return json.dumps(self.to_dict(), indent=4)

    def to_dict(self) -> dict:
        """
        Defines how a `valor.Evaluation` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing an evaluation.
        """
        return {
            "id": self.id,
            "model_name": self.model_name,
            "datum_filter": asdict(self.datum_filter),
            "parameters": asdict(self.parameters),
            "status": self.status.value,
            "metrics": self.metrics,
            "confusion_matrices": self.confusion_matrices,
            **self.kwargs,
        }

    def update(
        self,
        *_,
        id: int,
        model_name: str,
        datum_filter: Filter,
        parameters: EvaluationParameters,
        status: EvaluationStatus,
        metrics: List[Dict],
        confusion_matrices: List[Dict],
        **kwargs,
    ):
        self.id = id
        self.model_name = model_name
        self.datum_filter = (
            Filter(**datum_filter)
            if isinstance(datum_filter, dict)
            else datum_filter
        )
        self.parameters = (
            EvaluationParameters(**parameters)
            if isinstance(parameters, dict)
            else parameters
        )
        self.status = EvaluationStatus(status)
        self.metrics = metrics
        self.confusion_matrices = confusion_matrices
        self.kwargs = kwargs
        self.ignored_pred_labels: Optional[List[Label]] = None
        self.missing_pred_labels: Optional[List[Label]] = None

        for k, v in kwargs.items():
            setattr(self, k, v)

    def poll(self) -> EvaluationStatus:
        """
        Poll the back end.

        Updates the evaluation with the latest state from the back end.

        Returns
        -------
        enums.EvaluationStatus
            The status of the evaluation.

        Raises
        ----------
        ClientException
            If an Evaluation with the given `evaluation_id` is not found.
        """
        response = self.conn.get_evaluations(evaluation_ids=[self.id])
        if not response:
            raise ClientException("Not Found")
        self.update(**response[0])
        return self.status

    def wait_for_completion(
        self,
        *,
        timeout: Optional[int] = None,
        interval: float = 1.0,
    ) -> EvaluationStatus:
        """
        Blocking function that waits for evaluation to finish.

        Parameters
        ----------
        timeout : int, optional
            Length of timeout in seconds.
        interval : float, default=1.0
            Polling interval in seconds.
        """
        t_start = time.time()
        while self.poll() not in [
            EvaluationStatus.DONE,
            EvaluationStatus.FAILED,
        ]:
            time.sleep(interval)
            if timeout and time.time() - t_start > timeout:
                raise TimeoutError
        return self.status

    def to_dataframe(
        self,
        stratify_by: Optional[Tuple[str, str]] = None,
    ):
        """
        Get all metrics associated with a Model and return them in a `pd.DataFrame`.

        Returns
        ----------
        pd.DataFrame
            Evaluation metrics being displayed in a `pd.DataFrame`.

        Raises
        ------
        ModuleNotFoundError
            This function requires the use of `pandas.DataFrame`.

        """
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Must have pandas installed to use `get_metric_dataframes`."
            )

        if not stratify_by:
            column_type = "evaluation"
            column_name = self.id
        else:
            column_type = stratify_by[0]
            column_name = stratify_by[1]

        metrics = [
            {**metric, column_type: column_name} for metric in self.metrics
        ]
        df = pd.DataFrame(metrics)
        for k in ["label", "parameters"]:
            df[k] = df[k].fillna("n/a")
        df["parameters"] = df["parameters"].apply(json.dumps)
        df["label"] = df["label"].apply(
            lambda x: f"{x['key']}: {x['value']}" if x != "n/a" else x
        )
        df = df.pivot(
            index=["type", "parameters", "label"], columns=[column_type]
        )
        return df


@dataclass
class DatasetSummary:
    """Dataclass for storing dataset summary information"""

    name: str
    num_datums: int
    num_annotations: int
    num_bounding_boxes: int
    num_polygons: int
    num_rasters: int
    task_types: List[TaskType]
    labels: List[Label]
    datum_metadata: List[MetadataType]
    annotation_metadata: List[MetadataType]

    def __post_init__(self):
        for i, tt in enumerate(self.task_types):
            if isinstance(tt, str):
                self.task_types[i] = TaskType(tt)
        for i, label in enumerate(self.labels):
            if isinstance(label, dict):
                self.labels[i] = Label(**label)


class Dataset:
    """
    A class describing a given dataset.

    Parameters
    ----------
    name : str
        The name of the dataset.
    metadata : dict
        A dictionary of metadata that describes the dataset.
    """

    name = StringProperty("dataset_names")
    metadata = DictionaryProperty("dataset_metadata")

    def __init__(
        self,
        name: str,
        metadata: Optional[MetadataType] = None,
        connection: Optional[ClientConnection] = None,
    ):
        """
        Create or get a `Dataset` object.

        Parameters
        ----------
        name : str
            The name of the dataset.
        metadata : dict
            An optional dictionary of metadata that describes the dataset.
        connection : ClientConnnetion
            An optional Valor client object for interacting with the API.
        """
        self.conn = connection
        self.name = name
        self.metadata = metadata if metadata else {}

        # validation
        if not isinstance(self.name, str):
            raise TypeError("`name` should be of type `str`")
        validate_metadata(self.metadata)

    @classmethod
    def create(
        cls,
        name: str,
        metadata: Optional[MetadataType] = None,
    ) -> Dataset:
        """
        Creates a dataset that persists in the back end.

        Parameters
        ----------
        name : str
            The name of the dataset.
        metadata : dict
            A dictionary of metadata that describes the dataset.

        Returns
        -------
        valor.Dataset
            The created dataset.
        """
        dataset = cls(
            name=name,
            metadata=metadata,
        )
        Client(dataset.conn).create_dataset(dataset)
        return dataset

    @classmethod
    def get(
        cls,
        name: str,
    ) -> Union[Dataset, None]:
        """
        Retrieves a dataset from the back end database.

        Parameters
        ----------
        name : str
            The name of the dataset.

        Returns
        -------
        Union[valor.Dataset, None]
            The dataset or 'None' if it doesn't exist.
        """
        return Client().get_dataset(name)

    @classmethod
    def from_dict(
        cls, resp: dict, connection: Optional[ClientConnection] = None
    ) -> Dataset:
        """
        Construct a dataset from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing a dataset.
        connection : ClientConnection, optional
            Option to share a ClientConnection rather than request a new one.

        Returns
        -------
        valor.Dataset
        """
        resp.pop("id")
        resp["metadata"] = load_metadata(resp["metadata"])
        return cls(**resp, connection=connection)

    def to_dict(self, id: Optional[int] = None) -> dict:
        """
        Defines how a `valor.Dataset` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing a model.
        """
        return {
            "id": id,
            "name": self.name,
            "metadata": dump_metadata(self.metadata),
        }

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        objdict = self.to_dict()
        objdict.pop("id")
        return json.dumps(objdict, indent=4)

    def add_groundtruth(
        self,
        groundtruth: GroundTruth,
    ) -> None:
        """
        Add a ground truth to the dataset.

        Parameters
        ----------
        groundtruth : GroundTruth
            The ground truth to create.
        """
        Client(self.conn).create_groundtruths(
            dataset=self,
            groundtruths=[groundtruth],
        )

    def add_groundtruths(
        self,
        groundtruths: List[GroundTruth],
    ) -> None:
        """
        Add multiple ground truths to the dataset.

        Parameters
        ----------
        groundtruths : List[GroundTruth]
            The ground truths to create.
        """
        Client(self.conn).create_groundtruths(
            dataset=self,
            groundtruths=groundtruths,
        )

    def get_groundtruth(
        self,
        datum: Union[Datum, str],
    ) -> Union[GroundTruth, None]:
        """
        Get a particular ground truth.

        Parameters
        ----------
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[GroundTruth, None]
            The matching ground truth or 'None' if it doesn't exist.
        """
        return Client(self.conn).get_groundtruth(dataset=self, datum=datum)

    def get_labels(
        self,
    ) -> List[Label]:
        """
        Get all labels associated with a given dataset.

        Returns
        ----------
        List[Label]
            A list of `Labels` associated with the dataset.
        """
        return Client(self.conn).get_labels_from_dataset(self)

    def get_datums(
        self, filter_by: Optional[FilterType] = None
    ) -> List[Datum]:
        """
        Get all datums associated with a given dataset.

        Parameters
        ----------
        filter_by
            Optional constraints to filter by.

        Returns
        ----------
        List[Datum]
            A list of `Datums` associated with the dataset.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)

        if filter_.get("dataset_names"):
            raise ValueError(
                "Cannot filter by dataset_names when calling `Dataset.get_datums`."
            )
        filter_["dataset_names"] = [self.name]
        return Client(self.conn).get_datums(filter_by=filter_)

    def get_evaluations(
        self,
    ) -> List[Evaluation]:
        """
        Get all evaluations associated with a given dataset.

        Returns
        ----------
        List[Evaluation]
            A list of `Evaluations` associated with the dataset.
        """
        return Client(self.conn).get_evaluations(datasets=[self])

    def get_summary(self) -> DatasetSummary:
        """
        Get the summary of a given dataset.

        Returns
        -------
        DatasetSummary
            The summary of the dataset. This class has the following fields:

            name: name of the dataset

            num_datums: total number of datums in the dataset

            num_annotations: total number of labeled annotations in the dataset; if an
            object (such as a bounding box) has multiple labels, then each label is counted separately

            num_bounding_boxes: total number of bounding boxes in the dataset

            num_polygons: total number of polygons in the dataset

            num_rasters: total number of rasters in the dataset

            task_types: list of the unique task types in the dataset

            labels: list of the unique labels in the dataset

            datum_metadata: list of the unique metadata dictionaries in the dataset that are associated
            to datums

            groundtruth_annotation_metadata: list of the unique metadata dictionaries in the dataset that are
            associated to annotations
        """
        return Client(self.conn).get_dataset_summary(self.name)

    def finalize(
        self,
    ):
        """
        Finalizes the dataset such that new ground truths cannot be added to it.
        """
        return Client(self.conn).finalize_dataset(self)

    def delete(
        self,
        timeout: int = 0,
    ):
        """
        Delete the dataset from the back end.

        Parameters
        ----------
        timeout : int, default=0
            Sets a timeout in seconds.
        """
        Client(self.conn).delete_dataset(self.name, timeout)


class Model:
    """
    A class describing a model that was trained on a particular dataset.

    Parameters
    ----------
    name : str
        The name of the model.
    metadata : dict
        A dictionary of metadata that describes the model.
    """

    name = StringProperty("model_names")
    metadata = DictionaryProperty("model_metadata")

    def __init__(
        self,
        name: str,
        metadata: Optional[MetadataType] = None,
        connection: Optional[ClientConnection] = None,
    ):
        """
        Create or get a `Model` object.

        Parameters
        ----------
        name : str
            The name of the model.
        metadata : dict
            An optional dictionary of metadata that describes the dataset.
        connection : ClientConnnetion
            An optional Valor client object for interacting with the API.
        """
        self.conn = connection
        self.name = name
        self.metadata = metadata if metadata else {}

        # validation
        if not isinstance(self.name, str):
            raise TypeError("`name` should be of type `str`")
        validate_metadata(self.metadata)

    @classmethod
    def create(
        cls,
        name: str,
        metadata: Optional[MetadataType] = None,
    ) -> Model:
        """
        Creates a model that persists in the back end.

        Parameters
        ----------
        name : str
            The name of the model.
        metadata : dict
            A dictionary of metadata that describes the model.

        Returns
        -------
        valor.Model
            The created model.
        """
        model = cls(
            name=name,
            metadata=metadata,
        )
        Client(model.conn).create_model(model)
        return model

    @classmethod
    def get(
        cls,
        name: str,
    ) -> Union[Model, None]:
        """
        Retrieves a model from the back end database.

        Parameters
        ----------
        name : str
            The name of the model.

        Returns
        -------
        Union[valor.Model, None]
            The model or 'None' if it doesn't exist.
        """
        return Client().get_model(name)

    @classmethod
    def from_dict(
        cls, resp: dict, connection: Optional[ClientConnection] = None
    ) -> Model:
        """
        Construct a model from a dictionary.

        Parameters
        ----------
        resp : dict
            The dictionary containing a model.
        connection : ClientConnection, optional
            Option to share a ClientConnection rather than request a new one.

        Returns
        -------
        valor.Model
        """
        resp.pop("id")
        resp["metadata"] = load_metadata(resp["metadata"])
        return cls(**resp, connection=connection)

    def to_dict(self, id: Optional[int] = None) -> dict:
        """
        Defines how a `valor.Model` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing a model.
        """
        return {
            "id": id,
            "name": self.name,
            "metadata": dump_metadata(self.metadata),
        }

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        objdict = self.to_dict()
        objdict.pop("id")
        return json.dumps(objdict, indent=4)

    def add_prediction(
        self,
        dataset: Dataset,
        prediction: Prediction,
    ) -> None:
        """
        Add a prediction to the model.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        prediction : valor.Prediction
            The prediction to create.
        """
        Client(self.conn).create_predictions(
            dataset=dataset,
            model=self,
            predictions=[prediction],
        )

    def add_predictions(
        self,
        dataset: Dataset,
        predictions: List[Prediction],
    ) -> None:
        """
        Add multiple predictions to the model.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        predictions : List[valor.Prediction]
            The predictions to create.
        """
        Client(self.conn).create_predictions(
            dataset=dataset,
            model=self,
            predictions=predictions,
        )

    def get_prediction(
        self, dataset: Union[Dataset, str], datum: Union[Datum, str]
    ) -> Union[Prediction, None]:
        """
        Get a particular prediction.

        Parameters
        ----------
        dataset: Union[Dataset, str]
            The dataset the datum belongs to.
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[Prediction, None]
            The matching prediction or 'None' if it doesn't exist.
        """
        return Client(self.conn).get_prediction(
            dataset=dataset, model=self, datum=datum
        )

    def finalize_inferences(self, dataset: Union[Dataset, str]) -> None:
        """
        Finalizes the model over a dataset such that new predictions cannot be added to it.
        """
        if isinstance(dataset, Dataset):
            dataset = dataset.name
        return Client(self.conn).finalize_inferences(
            dataset=dataset, model=self
        )

    def _format_constraints(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
    ) -> Filter:
        """Formats the 'datum_filter' for any evaluation requests."""

        # get list of dataset names
        dataset_names_from_obj = []
        if isinstance(datasets, list):
            dataset_names_from_obj = [dataset.name for dataset in datasets]
        elif isinstance(datasets, Dataset):
            dataset_names_from_obj = [datasets.name]

        # create a 'schemas.Filter' object from the constraints.
        filter_ = _format_filter(filter_by)

        # reset model name
        filter_.model_names = None
        filter_.model_metadata = None

        # set dataset names
        if not filter_.dataset_names:
            filter_.dataset_names = []
        filter_.dataset_names.extend(dataset_names_from_obj)
        return filter_

    def _create_label_map(
        self,
        label_map: Optional[Dict[Label, Label]],
    ) -> Union[List[List[List[str]]], None]:
        """Convert a dictionary of label maps to a serializable list format."""
        if not label_map:
            return None

        if not isinstance(label_map, dict) or not all(
            [
                isinstance(key, Label) and isinstance(value, Label)
                for key, value in label_map.items()
            ]
        ):
            raise TypeError(
                "label_map should be a dictionary with valid Labels for both the key and value."
            )

        return_value = [
            [[key.key, key.value], [value.key, value.value]]
            for key, value in label_map.items()
        ]

        return return_value

    def evaluate_classification(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        compute_pr_curves: bool = False,
    ) -> Evaluation:
        """
        Start a classification evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]], optional
            The dataset or list of datasets to evaluate against.
        filter_by : FilterType, optional
            Optional set of constraints to filter evaluation by.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        compute_pr_curves: bool
            A boolean which determines whether we calculate precision-recall curves or not.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if not datasets and not filter_by:
            raise ValueError(
                "Evaluation requires the definition of either datasets, dataset filters or both."
            )

        # format request
        datum_filter = self._format_constraints(datasets, filter_by)
        request = EvaluationRequest(
            model_names=[self.name],
            datum_filter=datum_filter,
            parameters=EvaluationParameters(
                task_type=TaskType.CLASSIFICATION,
                label_map=self._create_label_map(label_map=label_map),
                compute_pr_curves=compute_pr_curves,
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_detection(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
        convert_annotations_to_type: Optional[AnnotationType] = None,
        iou_thresholds_to_compute: Optional[List[float]] = None,
        iou_thresholds_to_return: Optional[List[float]] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        recall_score_threshold: float = 0,
        compute_pr_curves: bool = False,
        pr_curve_iou_threshold: float = 0.5,
    ) -> Evaluation:
        """
        Start an object-detection evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]], optional
            The dataset or list of datasets to evaluate against.
        filter_by : FilterType, optional
            Optional set of constraints to filter evaluation by.
        convert_annotations_to_type : enums.AnnotationType, optional
            Forces the object detection evaluation to compute over this type.
        iou_thresholds_to_compute : List[float], optional
            Thresholds to compute mAP against.
        iou_thresholds_to_return : List[float], optional
            Thresholds to return AP for. Must be subset of `iou_thresholds_to_compute`.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        recall_score_threshold: float, default=0
            The confidence score threshold for use when determining whether to count a prediction as a true positive or not while calculating Average Recall.
        compute_pr_curves: bool, optional
            A boolean which determines whether we calculate precision-recall curves or not.
        pr_curve_iou_threshold: float, optional
            The IOU threshold to use when calculating precision-recall curves. Defaults to 0.5. Does nothing when compute_pr_curves is set to False or None.


        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if iou_thresholds_to_compute is None:
            iou_thresholds_to_compute = [
                round(0.5 + 0.05 * i, 2) for i in range(10)
            ]
        if iou_thresholds_to_return is None:
            iou_thresholds_to_return = [0.5, 0.75]

        # format request
        parameters = EvaluationParameters(
            task_type=TaskType.OBJECT_DETECTION,
            convert_annotations_to_type=convert_annotations_to_type,
            iou_thresholds_to_compute=iou_thresholds_to_compute,
            iou_thresholds_to_return=iou_thresholds_to_return,
            label_map=self._create_label_map(label_map=label_map),
            recall_score_threshold=recall_score_threshold,
            compute_pr_curves=compute_pr_curves,
            pr_curve_iou_threshold=pr_curve_iou_threshold,
        )
        datum_filter = self._format_constraints(datasets, filter_by)
        request = EvaluationRequest(
            model_names=[self.name],
            datum_filter=datum_filter,
            parameters=parameters,
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_segmentation(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
        label_map: Optional[Dict[Label, Label]] = None,
    ) -> Evaluation:
        """
        Start a semantic-segmentation evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]], optional
            The dataset or list of datasets to evaluate against.
        filter_by : FilterType, optional
            Optional set of constraints to filter evaluation by.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion
        """
        # format request
        datum_filter = self._format_constraints(datasets, filter_by)
        request = EvaluationRequest(
            model_names=[self.name],
            datum_filter=datum_filter,
            parameters=EvaluationParameters(
                task_type=TaskType.SEMANTIC_SEGMENTATION,
                label_map=self._create_label_map(label_map=label_map),
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def delete(self, timeout: int = 0):
        """
        Delete the `Model` object from the back end.

        Parameters
        ----------
        timeout : int, default=0
            Sets a timeout in seconds.
        """
        Client(self.conn).delete_model(self.name, timeout)

    def get_labels(
        self,
    ) -> List[Label]:
        """
        Get all labels associated with a given model.

        Returns
        ----------
        List[Label]
            A list of `Labels` associated with the model.
        """
        return Client(self.conn).get_labels_from_model(self)

    def get_evaluations(
        self,
    ) -> List[Evaluation]:
        """
        Get all evaluations associated with a given model.

        Returns
        ----------
        List[Evaluation]
            A list of `Evaluations` associated with the model.
        """
        return Client(self.conn).get_evaluations(models=[self])


class Client:
    """
    Valor client object for interacting with the api.

    Parameters
    ----------
    connection : ClientConnection, optional
        Option to use an existing connection object.
    """

    def __init__(self, connection: Optional[ClientConnection] = None):
        if not connection:
            connection = get_connection()
        self.conn = connection

    @classmethod
    def connect(
        cls,
        host: str,
        access_token: Optional[str] = None,
        reconnect: bool = False,
    ) -> Client:
        """
        Establishes a connection to the Valor API.

        Parameters
        ----------
        host : str
            The host to connect to. Should start with "http://" or "https://".
        access_token : str
            The access token for the host (if the host requires authentication).
        """
        connect(host=host, access_token=access_token, reconnect=reconnect)
        return cls(get_connection())

    def get_labels(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Label]:
        """
        Gets all labels using an optional filter.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        filter_ = _format_filter(filter_by)
        filter_ = asdict(filter_)
        return [
            Label.from_dict(label) for label in self.conn.get_labels(filter_)
        ]

    def get_labels_from_dataset(
        self, dataset: Union[Dataset, str]
    ) -> List[Label]:
        """
        Get all labels associated with a dataset's ground truths.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset to search by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        return [
            Label.from_dict(label)
            for label in self.conn.get_labels_from_dataset(dataset_name)
        ]

    def get_labels_from_model(self, model: Union[Model, str]) -> List[Label]:
        """
        Get all labels associated with a model's ground truths.

        Parameters
        ----------
        model : valor.Model
            The model to search by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        model_name = model.name if isinstance(model, Model) else model
        return [
            Label.from_dict(label)
            for label in self.conn.get_labels_from_model(model_name)
        ]

    def create_dataset(
        self,
        dataset: Union[Dataset, dict],
    ) -> None:
        """
        Creates a dataset.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset to create.
        """
        if isinstance(dataset, Dataset):
            dataset = dataset.to_dict()
        self.conn.create_dataset(dataset)

    def create_groundtruths(
        self,
        dataset: Dataset,
        groundtruths: List[GroundTruth],
    ):
        """
        Creates ground truths.

        Parameters
        ----------

        dataset : valor.Dataset
            The dataset to create the ground truth for.
        groundtruths : List[valor.GroundTruth]
            The ground truths to create.
        """
        groundtruths_json = []
        for groundtruth in groundtruths:
            if not isinstance(groundtruth, GroundTruth):
                raise TypeError(
                    f"Expected ground truth to be of type 'valor.GroundTruth' not '{type(groundtruth)}'."
                )
            if len(groundtruth.annotations) == 0:
                warnings.warn(
                    f"GroundTruth for datum with uid `{groundtruth.datum.uid}` contains no annotations."
                )
            groundtruths_json.append(
                groundtruth.to_dict(dataset_name=dataset.name)
            )
        self.conn.create_groundtruths(groundtruths_json)

    def get_groundtruth(
        self,
        dataset: Union[Dataset, str],
        datum: Union[Datum, str],
    ) -> Union[GroundTruth, None]:
        """
        Get a particular ground truth.

        Parameters
        ----------
        dataset: Union[Dataset, str]
            The dataset the datum belongs to.
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[GroundTruth, None]
            The matching ground truth or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        datum_uid = datum.uid if isinstance(datum, Datum) else datum

        try:
            resp = self.conn.get_groundtruth(
                dataset_name=dataset_name, datum_uid=datum_uid
            )
            return GroundTruth.from_dict(resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def finalize_dataset(self, dataset: Union[Dataset, str]) -> None:
        """
        Finalizes a dataset such that new ground truths cannot be added to it.

        Parameters
        ----------
        dataset : str
            The dataset to be finalized.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        return self.conn.finalize_dataset(name=dataset_name)

    def get_dataset(
        self,
        name: str,
    ) -> Union[Dataset, None]:
        """
        Gets a dataset by name.

        Parameters
        ----------
        name : str
            The name of the dataset to fetch.

        Returns
        -------
        Union[Dataset, None]
            A Dataset with a matching name, or 'None' if one doesn't exist.
        """
        try:
            return Dataset.from_dict(
                self.conn.get_dataset(name), connection=self.conn
            )
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_datasets(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Dataset]:
        """
        Get all datasets, with an option to filter results according to some user-defined parameters.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Dataset]
            A list of datasets.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)
        return [
            Dataset.from_dict(dataset, connection=self.conn)
            for dataset in self.conn.get_datasets(filter_)
        ]

    def get_datums(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Datum]:
        """
        Get all datums using an optional filter.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        -------
        List[valor.Datum]
            A list datums.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)
        return [
            Datum.from_dict(datum) for datum in self.conn.get_datums(filter_)
        ]

    def get_datum(
        self,
        dataset: Union[Dataset, str],
        uid: str,
    ) -> Union[Datum, None]:
        """
        Get datum.
        `GET` endpoint.
        Parameters
        ----------
        dataset : valor.Dataset
            The dataset the datum belongs to.
        uid : str
            The UID of the datum.
        Returns
        -------
        valor.Datum
            The requested datum or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        try:
            resp = self.conn.get_datum(dataset_name=dataset_name, uid=uid)
            return Datum.from_dict(resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_dataset_status(
        self,
        name: str,
    ) -> Union[TableStatus, None]:
        """
        Get the state of a given dataset.

        Parameters
        ----------
        name : str
            The name of the dataset we want to fetch the state of.

        Returns
        ------
        TableStatus | None
            The state of the dataset, or 'None' if the dataset does not exist.
        """
        try:
            return self.conn.get_dataset_status(name)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_dataset_summary(self, name: str) -> DatasetSummary:
        """
        Gets the summary of a dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to create a summary for.

        Returns
        -------
        DatasetSummary
            A dataclass containing the dataset summary.
        """
        return DatasetSummary(**self.conn.get_dataset_summary(name))

    def delete_dataset(self, name: str, timeout: int = 0) -> None:
        """
        Deletes a dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to be deleted.
        timeout : int
            The number of seconds to wait in order to confirm that the dataset was deleted.
        """
        self.conn.delete_dataset(name)
        if timeout:
            for _ in range(timeout):
                if self.get_dataset(name) is None:
                    break
                else:
                    time.sleep(1)
            else:
                raise TimeoutError(
                    "Dataset wasn't deleted within timeout interval"
                )

    def create_model(
        self,
        model: Union[Model, dict],
    ):
        """
        Creates a model.

        Parameters
        ----------
        model : valor.Model
            The model to create.
        """
        if isinstance(model, Model):
            model = model.to_dict()
        self.conn.create_model(model)

    def create_predictions(
        self,
        dataset: Dataset,
        model: Model,
        predictions: List[Prediction],
    ) -> None:
        """
        Creates predictions.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        model : valor.Model
            The model making the prediction.
        predictions : List[valor.Prediction]
            The predictions to create.
        """
        predictions_json = []
        for prediction in predictions:
            if not isinstance(prediction, Prediction):
                raise TypeError(
                    f"Expected prediction to be of type 'valor.Prediction' not '{type(prediction)}'."
                )
            if len(prediction.annotations) == 0:
                warnings.warn(
                    f"Prediction for datum with uid `{prediction.datum.uid}` contains no annotations."
                )
            predictions_json.append(
                prediction.to_dict(
                    dataset_name=dataset.name,
                    model_name=model.name,
                )
            )
        self.conn.create_predictions(predictions_json)

    def get_prediction(
        self,
        dataset: Union[Dataset, str],
        model: Union[Model, str],
        datum: Union[Datum, str],
    ) -> Union[Prediction, None]:
        """
        Get a particular prediction.

        Parameters
        ----------
        dataset: Union[Dataset, str]
            The dataset the datum belongs to.
        model: Union[Model, str]
            The model that made the prediction.
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[Prediction, None]
            The matching prediction or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        model_name = model.name if isinstance(model, Model) else model
        datum_uid = datum.uid if isinstance(datum, Datum) else datum

        try:
            resp = self.conn.get_prediction(
                dataset_name=dataset_name,
                model_name=model_name,
                datum_uid=datum_uid,
            )
            return Prediction.from_dict(resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def finalize_inferences(
        self, dataset: Union[Dataset, str], model: Union[Model, str]
    ) -> None:
        """
        Finalizes a model-dataset pairing such that new predictions cannot be added to it.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        model_name = model.name if isinstance(model, Model) else model
        return self.conn.finalize_inferences(
            dataset_name=dataset_name,
            model_name=model_name,
        )

    def get_model(
        self,
        name: str,
    ) -> Union[Model, None]:
        """
        Gets a model by name.

        Parameters
        ----------
        name : str
            The name of the model to fetch.

        Returns
        -------
        Union[valor.Model, None]
            A Model with matching name or 'None' if one doesn't exist.
        """
        try:
            return Model.from_dict(
                self.conn.get_model(name), connection=self.conn
            )
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_models(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Model]:
        """
        Get all models using an optional filter.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Model]
            A list of models.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)
        return [
            Model.from_dict(model, connection=self.conn)
            for model in self.conn.get_models(filter_)
        ]

    def get_model_status(
        self,
        dataset_name: str,
        model_name: str,
    ) -> Optional[TableStatus]:
        """
        Get the state of a given model over a dataset.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset that the model is operating over.
        model_name : str
            The name of the model we want to fetch the state of.

        Returns
        ------
        Union[TableStatus, None]
            The state of the model or 'None' if the model doesn't exist.
        """
        try:
            return self.conn.get_model_status(dataset_name, model_name)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_model_eval_requests(
        self, model: Union[Model, str]
    ) -> List[Evaluation]:
        """
        Get all evaluations that have been created for a model.

        This does not return evaluation results.

        `GET` endpoint.

        Parameters
        ----------
        model : str
            The model to search by.

        Returns
        -------
        List[Evaluation]
            A list of evaluations.
        """
        model_name = model.name if isinstance(model, Model) else model
        return [
            Evaluation(**evaluation, connection=self.conn)
            for evaluation in self.conn.get_model_eval_requests(model_name)
        ]

    def delete_model(self, name: str, timeout: int = 0) -> None:
        """
        Deletes a model.

        Parameters
        ----------
        name : str
            The name of the model to be deleted.
        timeout : int
            The number of seconds to wait in order to confirm that the model was deleted.
        """
        self.conn.delete_model(name)
        if timeout:
            for _ in range(timeout):
                if self.get_model(name) is None:
                    break
                else:
                    time.sleep(1)
            else:
                raise TimeoutError(
                    "Model wasn't deleted within timeout interval"
                )

    def get_evaluations(
        self,
        *,
        evaluation_ids: Optional[List[int]] = None,
        models: Union[List[Model], List[str], None] = None,
        datasets: Union[List[Dataset], List[str], None] = None,
    ) -> List[Evaluation]:
        """
        Returns all evaluations associated with user-supplied dataset and/or model names.

        Parameters
        ----------
        evaluation_ids : List[int], optional.
            A list of job IDs to return metrics for.
        models : Union[List[valor.Model], List[str]], optional
            A list of model names that we want to return metrics for.
        datasets : Union[List[valor.Dataset], List[str]], optional
            A list of dataset names that we want to return metrics for.

        Returns
        -------
        List[valor.Evaluation]
            A list of evaluations.
        """
        if isinstance(datasets, list):
            datasets = [
                element.name if isinstance(element, Dataset) else element
                for element in datasets
            ]
        if isinstance(models, list):
            models = [
                element.name if isinstance(element, Model) else element
                for element in models
            ]
        return [
            Evaluation(connection=self.conn, **evaluation)
            for evaluation in self.conn.get_evaluations(
                evaluation_ids=evaluation_ids,
                models=models,
                datasets=datasets,
            )
        ]

    def evaluate(self, request: EvaluationRequest) -> List[Evaluation]:
        """
        Creates as many evaluations as necessary to fulfill the request.

        Parameters
        ----------
        request : schemas.EvaluationRequest
            The requested evaluation parameters.

        Returns
        -------
        List[Evaluation]
            A list of evaluations that meet the parameters.
        """
        return [
            Evaluation(**evaluation)
            for evaluation in self.conn.evaluate(request)
        ]



from dataclasses import dataclass
from typing import Dict, Iterable, Iterator, List, Optional, Union

from valor.enums import TaskType
from valor.schemas.constraints import BinaryExpression, Constraint


@dataclass
class Filter:
    """
    Used to filter Evaluations according to specific, user-defined criteria.

    Attributes
    ----------
    dataset_names : List[str], optional
        A list of `Dataset` names to filter on.
    dataset_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Dataset` metadata to filter on.
    model_names : List[str], optional
        A list of `Model` names to filter on.
    model_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Model` metadata to filter on.
    datum_uids : List[str], optional
        A list of `Datum` UIDs to filter on.
    datum_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Datum` metadata to filter on.
    task_types : List[TaskType], optional
        A list of task types to filter on.
    annotation_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Annotation` metadata to filter on.
    require_bounding_box : bool, optional
        A toggle for filtering by bounding boxes.
    bounding_box_area : bool, optional
        An optional constraint to filter by bounding box area.
    require_polygon : bool, optional
        A toggle for filtering by polygons.
    polygon_area : bool, optional
        An optional constraint to filter by polygon area.
    require_raster : bool, optional
        A toggle for filtering by rasters.
    raster_area : bool, optional
        An optional constraint to filter by raster area.
    labels : List[Label], optional
        A list of `Labels' to filter on.
    label_ids : List[int], optional
        A list of label row id's.
    label_keys : List[str], optional
        A list of `Label` keys to filter on.
    label_scores : List[Constraint], optional
        A list of `Constraints` which are used to filter `Evaluations` according to the `Model`'s prediction scores.

    Raises
    ------
    TypeError
        If `value` isn't of the correct type.
    ValueError
        If the `operator` doesn't match one of the allowed patterns.
    """

    # datasets
    dataset_names: Optional[List[str]] = None
    dataset_metadata: Optional[Dict[str, List[Constraint]]] = None

    # models
    model_names: Optional[List[str]] = None
    model_metadata: Optional[Dict[str, List[Constraint]]] = None

    # datums
    datum_uids: Optional[List[str]] = None
    datum_metadata: Optional[Dict[str, List[Constraint]]] = None

    # annotations
    task_types: Optional[List[TaskType]] = None
    annotation_metadata: Optional[Dict[str, List[Constraint]]] = None

    # geometries
    require_bounding_box: Optional[bool] = None
    bounding_box_area: Optional[List[Constraint]] = None
    require_polygon: Optional[bool] = None
    polygon_area: Optional[List[Constraint]] = None
    require_raster: Optional[bool] = None
    raster_area: Optional[List[Constraint]] = None

    # labels
    labels: Optional[List[Dict[str, str]]] = None
    label_ids: Optional[List[int]] = None
    label_keys: Optional[List[str]] = None
    label_scores: Optional[List[Constraint]] = None

    def __post_init__(self):
        def _unpack_metadata(metadata: Optional[dict]) -> Union[dict, None]:
            if metadata is None:
                return None
            for k, vlist in metadata.items():
                metadata[k] = [
                    v if isinstance(v, Constraint) else Constraint(**v)
                    for v in vlist
                ]
            return metadata

        # unpack metadata
        self.dataset_metadata = _unpack_metadata(self.dataset_metadata)
        self.model_metadata = _unpack_metadata(self.model_metadata)
        self.datum_metadata = _unpack_metadata(self.datum_metadata)
        self.annotation_metadata = _unpack_metadata(self.annotation_metadata)

        def _unpack_list(
            vlist: Optional[list], object_type: type
        ) -> Optional[list]:
            def _handle_conversion(v, object_type):
                if object_type is Constraint:
                    return object_type(**v)
                else:
                    return object_type(v)

            if vlist is None:
                return None

            return [
                (
                    v
                    if isinstance(v, object_type)
                    else _handle_conversion(v=v, object_type=object_type)
                )
                for v in vlist
            ]

        # unpack tasktypes
        self.task_types = _unpack_list(self.task_types, TaskType)

        # unpack area
        self.bounding_box_area = _unpack_list(
            self.bounding_box_area, Constraint
        )
        self.polygon_area = _unpack_list(self.polygon_area, Constraint)
        self.raster_area = _unpack_list(self.raster_area, Constraint)

        # scores
        self.label_scores = _unpack_list(self.label_scores, Constraint)

    @classmethod
    def create(
        cls, expressions: List[Union[BinaryExpression, List[BinaryExpression]]]
    ):
        """
        Parses a list of `BinaryExpression` to create a `schemas.Filter` object.

        Parameters
        ----------
        expressions: Sequence[Union[BinaryExpression, Sequence[BinaryExpression]]]
            A list of (lists of) `BinaryExpressions' to parse into a `Filter` object.
        """

        def flatten(
            t: Iterable[Union[BinaryExpression, Iterable[BinaryExpression]]]
        ) -> Iterator[BinaryExpression]:
            """Flatten a nested iterable of BinaryExpressions."""
            for item in t:
                if isinstance(item, BinaryExpression):
                    yield item
                else:
                    yield from flatten(item)

        # create dict using expr names as keys
        expression_dict = {}
        for expr in flatten(expressions):
            if expr.name not in expression_dict:
                expression_dict[expr.name] = []
            expression_dict[expr.name].append(expr)

        # create filter
        filter_request = cls()

        # metadata constraints
        for attr in [
            "dataset_metadata",
            "model_metadata",
            "datum_metadata",
            "annotation_metadata",
        ]:
            if attr in expression_dict:
                for expr in expression_dict[attr]:
                    if not getattr(filter_request, attr):
                        setattr(filter_request, attr, {})
                    __value = getattr(filter_request, attr)
                    if expr.key not in __value:
                        __value[expr.key] = []
                    __value[expr.key].append(expr.constraint)
                    setattr(filter_request, attr, __value)

        # numeric constraints
        for attr in [
            "bounding_box_area",
            "polygon_area",
            "raster_area",
            "label_scores",
        ]:
            if attr in expression_dict:
                setattr(
                    filter_request,
                    attr,
                    [expr.constraint for expr in expression_dict[attr]],
                )

        # boolean constraints
        for attr in [
            "require_bounding_box",
            "require_polygon",
            "require_raster",
        ]:
            if attr in expression_dict:
                for expr in expression_dict[attr]:
                    if expr.constraint.operator == "exists":
                        setattr(filter_request, attr, True)
                    elif expr.constraint.operator == "is_none":
                        setattr(filter_request, attr, False)

        # equality constraints
        for attr in [
            "dataset_names",
            "model_names",
            "datum_uids",
            "task_types",
            "labels",
            "label_keys",
        ]:
            if attr in expression_dict:
                setattr(
                    filter_request,
                    attr,
                    [expr.constraint.value for expr in expression_dict[attr]],
                )

        return filter_request



import numpy as np
import PIL.Image
import pytest

from valor import Annotation, GroundTruth, Label
from valor.enums import TaskType
from valor.metatypes import ImageMetadata
from valor.schemas import BasicPolygon, MultiPolygon, Point, Polygon, Raster
from valor.viz import (
    _polygons_to_binary_mask,
    create_combined_segmentation_mask,
    draw_detections_on_image,
)


@pytest.fixture
def bounding_poly() -> BasicPolygon:
    return BasicPolygon(
        points=[
            Point(100, 100),
            Point(200, 100),
            Point(200, 200),
            Point(100, 200),
        ]
    )


@pytest.fixture
def poly1(bounding_poly: BasicPolygon) -> Polygon:
    return Polygon(
        boundary=bounding_poly,
        holes=[
            BasicPolygon(
                points=[
                    Point(150, 120),
                    Point(180, 120),
                    Point(180, 140),
                    Point(150, 140),
                ]
            )
        ],
    )


def test__polygons_to_binary_mask(poly1):
    poly2 = Polygon(
        boundary=BasicPolygon(
            points=[
                Point(10, 15),
                Point(20, 15),
                Point(20, 20),
                Point(10, 20),
            ]
        )
    )

    mask = _polygons_to_binary_mask([poly1, poly2], 500, 600)

    area_poly1 = (200 - 100 + 1) * (200 - 100 + 1) - (180 - 150 + 1) * (
        140 - 120 + 1
    )

    area_poly2 = (20 - 10 + 1) * (20 - 15 + 1)

    assert mask.sum() == area_poly1 + area_poly2


def test_create_combined_segmentation_mask(poly1: Polygon):
    with pytest.raises(ValueError) as exc_info:
        create_combined_segmentation_mask(
            [],
            label_key="",
            task_type=TaskType.OBJECT_DETECTION,
        )
    assert "cannot be empty" in str(exc_info)

    image = ImageMetadata(uid="uid", height=200, width=200).to_datum()

    gt1 = GroundTruth(
        datum=image,
        annotations=[
            Annotation(
                task_type=TaskType.OBJECT_DETECTION,
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v2"),
                    Label(key="k3", value="v3"),
                ],
                raster=Raster.from_geometry(
                    MultiPolygon(polygons=[poly1]),
                    height=200,
                    width=200,
                ),
            ),
        ],
    )

    gt2 = GroundTruth(
        datum=image,
        annotations=[
            Annotation(
                task_type=TaskType.SEMANTIC_SEGMENTATION,
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v3"),
                ],
                raster=Raster.from_numpy(
                    np.array([[True, False], [False, True]]),
                ),
            )
        ],
    )

    gts = [gt1, gt2]

    # check get an error since "k3" isn't a label key in seg2
    with pytest.raises(RuntimeError) as exc_info:
        create_combined_segmentation_mask(
            [gts[1]],
            label_key="k3",
            task_type=TaskType.SEMANTIC_SEGMENTATION,
        )
    assert "doesn't have a label" in str(exc_info)

    # should have one distinct (non-black) color
    combined_mask, _ = create_combined_segmentation_mask(
        gts,
        label_key="k1",
        task_type=TaskType.SEMANTIC_SEGMENTATION,
    )
    combined_mask = np.array(combined_mask)
    # check that we get two unique RGB values (black and one color for label value "v1")
    unique_rgb = np.unique(combined_mask.reshape(-1, 3), axis=0)
    assert unique_rgb.shape == (2, 3)

    # should have two distinct (non-black) color
    combined_mask, _ = create_combined_segmentation_mask(gts, label_key="k2")
    combined_mask = np.array(combined_mask)
    # check that we get two unique RGB values (black and one color for label value "v1")
    unique_rgb = np.unique(combined_mask.reshape(-1, 3), axis=0)
    assert unique_rgb.shape == (3, 3)

    with pytest.raises(RuntimeError) as exc_info:
        create_combined_segmentation_mask(
            gts
            + [
                GroundTruth(
                    datum=ImageMetadata(
                        "different uid", height=10, width=100
                    ).to_datum(),
                    annotations=gts[0].annotations,
                )
            ],
            "",
        )
    assert "belong to the same image" in str(exc_info)


def test_draw_detections_on_image(bounding_poly: BasicPolygon):
    detections = [
        GroundTruth(
            datum=ImageMetadata("test", 300, 300).to_datum(),
            annotations=[
                Annotation(
                    task_type=TaskType.OBJECT_DETECTION,
                    labels=[Label("k", "v")],
                    polygon=Polygon(
                        boundary=bounding_poly,
                    ),
                )
            ],
        ),
    ]
    img = PIL.Image.new("RGB", (300, 300))

    img = draw_detections_on_image(detections, img)

    assert img.size == (300, 300)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(img).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0



import io
from base64 import b64decode, b64encode
from dataclasses import dataclass, field
from typing import List, Tuple, Union

import numpy as np
import PIL.Image


@dataclass
class Point:
    """
    Represents a point in 2D space.

    Parameters
    ----------
    x : Union[float, int]
        The x-coordinate of the point.
    y : Union[float, int]
        The y-coordinate of the point.

    Attributes
    ----------
    x : float
        The x-coordinate of the point.
    y : float
        The y-coordinate of the point.

    Raises
    ------
    TypeError
        If the coordinates are not of type `float` or convertible to `float`.

    Examples
    --------
    >>> Point(1,2)
    Point(x=1.0, y=2.0)
    """

    x: float
    y: float

    def __post_init__(self):
        self.x = float(self.x)
        self.y = float(self.y)

    def __hash__(self):
        return hash(f"{self.x},{self.y}")

    def resize(
        self,
        og_img_h: int,
        og_img_w: int,
        new_img_h: int,
        new_img_w: int,
    ) -> "Point":
        """
        Resize the point coordinates based on the scaling factors.

        Parameters
        ----------
        og_img_h : int
            Original image height.
        og_img_w : int
            Original image width.
        new_img_h : int
            New image height.
        new_img_w : int
            New image width.

        Returns
        -------
        Point
            Resized point based on the scaling factors.

        Examples
        --------
        >>> p = Point(1,2)
        >>> p.resize(og_img_h=100,
        ...          og_img_w=50,
        ...          new_img_h=1000,
        ...          new_img_w=1000)
        Point(x=20.0, y=20.0)
        """
        h_factor, w_factor = new_img_h / og_img_h, new_img_w / og_img_w
        return Point(x=w_factor * self.x, y=h_factor * self.y)


@dataclass
class BasicPolygon:
    """
    Class for representing a bounding region.

    Attributes
    ----------
    points : List[Point], optional
        List of `Point` objects representing the vertices of the polygon. Defaults to an empty list.

    Raises
    ------
    TypeError
        If `points` is not a list, or an element in `points` is not a `Point`.
    ValueError
        If the number of unique points in `points` is less than 3,
        making the BasicPolygon invalid.

    Examples
    --------
    >>> BasicPolygon(
    ...     points=[
    ...         Point(0,0),
    ...         Point(0,1),
    ...         Point(1,0),
    ...     ]
    ... )
    """

    points: List[Point] = field(default_factory=list)

    def __post_init__(self):
        # unpack & validate
        if not isinstance(self.points, list):
            raise TypeError("Member `points` is not a list.")
        for i, point in enumerate(self.points):
            if isinstance(point, dict):
                self.points[i] = Point(**point)
            if not isinstance(self.points[i], Point):
                raise TypeError("Element in points is not a `Point`.")
        if len(set(self.points)) < 3:
            raise ValueError(
                "BasicPolygon needs at least 3 unique points to be valid."
            )

    def xy_list(self) -> List[Point]:
        """
        Returns a list of `Point` objects representing the vertices of the polygon.

        Returns
        -------
        List[Point]
            List of `Point` objects.
        """
        return self.points.copy()

    def tuple_list(self) -> List[Tuple[float, float]]:
        """
        Returns a list of points as tuples (x, y).

        Returns
        -------
        List[Tuple[float, float]]
            List of points as tuples.
        """
        return [(pt.x, pt.y) for pt in self.points]

    @property
    def xmin(self):
        """Minimum x-coordinate of the polygon."""
        return min(p.x for p in self.points)

    @property
    def ymin(self):
        """Minimum y-coordinate of the polygon."""
        return min(p.y for p in self.points)

    @property
    def xmax(self):
        """Maximum x-coordinate of the polygon."""
        return max(p.x for p in self.points)

    @property
    def ymax(self):
        """Maximum y-coordinate of the polygon."""
        return max(p.y for p in self.points)


@dataclass
class Polygon:
    """
    Represents a polygon with a boundary and optional holes.

    Parameters
    ----------
    boundary : BasicPolygon or dict
        The outer boundary of the polygon. Can be a `BasicPolygon` object or a
        dictionary with the necessary information to create a `BasicPolygon`.
    holes : List[BasicPolygon], optional
        List of holes inside the polygon. Defaults to an empty list.

    Raises
    ------
    TypeError
        If `boundary` is not a `BasicPolygon` or cannot be converted to one.
        If `holes` is not a list, or an element in `holes` is not a `BasicPolygon`.

    Examples
    --------
    Create component polygons with BasicPolygon
    >>> basic_polygon1 = BasicPolygon(...)
    >>> basic_polygon2 = BasicPolygon(...)
    >>> basic_polygon3 = BasicPolygon(...)

    Create a polygon from a basic polygon.
    >>> polygon1 = Polygon(
    ...     boundary=basic_polygon1,
    ... )

    Create a polygon with holes.
    >>> polygon2 = Polygon(
    ...     boundary=basic_polygon1,
    ...     holes=[basic_polygon2, basic_polygon3],
    ... )
    """

    boundary: BasicPolygon
    holes: List[BasicPolygon] = field(default_factory=list)

    def __post_init__(self):
        # unpack & validate
        if isinstance(self.boundary, dict):
            self.boundary = BasicPolygon(**self.boundary)
        if not isinstance(self.boundary, BasicPolygon):
            raise TypeError(
                "boundary should be of type `valor.schemas.BasicPolygon`"
            )
        if self.holes:
            if not isinstance(self.holes, list):
                raise TypeError(
                    f"holes should be a list of `valor.schemas.BasicPolygon`. Got `{type(self.holes)}`."
                )
            for i, hole in enumerate(self.holes):
                if isinstance(hole, dict):
                    self.holes[i] = BasicPolygon(**hole)
                if not isinstance(self.holes[i], BasicPolygon):
                    raise TypeError(
                        "holes list should contain elements of type `valor.schemas.BasicPolygon`"
                    )
        else:
            self.holes = []


@dataclass
class BoundingBox:
    """
    Represents a bounding box defined by a 4-point polygon. Note that this does not need to be axis-aligned.

    Parameters
    ----------
    polygon : BasicPolygon or dict
        The 4-point polygon defining the bounding box. Can be a `BasicPolygon` object
        or a dictionary with the necessary information to create a `BasicPolygon`.

    Raises
    ------
    TypeError
        If `polygon` is not a `BasicPolygon` or cannot be converted to one.
    ValueError
        If the number of points in `polygon` is not equal to 4, making it invalid as a bounding box.

    Examples
    --------
    Create a BoundingBox from Points.
    Note that ordering is important to prevent self-intersection!
    >>> box1 = schemas.BoundingBox(
    ...     polygon=schemas.BasicPolygon(
    ...         points=[
    ...             schemas.Point(0,0),
    ...             schemas.Point(0,1),
    ...             schemas.Point(1,1),
    ...             schemas.Point(1,0),
    ...         ]
    ...     ),
    ... )

    Create a BoundingBox using extrema.
    >>> box2 = BoundingBox.from_extrema(
    ...     xmin=0, xmax=1,
    ...     ymin=0, ymax=1,
    ... )
    """

    polygon: BasicPolygon

    def __post_init__(self):
        if isinstance(self.polygon, dict):
            self.polygon = BasicPolygon(**self.polygon)
        if not isinstance(self.polygon, BasicPolygon):
            raise TypeError(
                "polygon should be of type `valor.schemas.BasicPolygon`"
            )
        if len(self.polygon.points) != 4:
            raise ValueError(
                "Bounding box should be made of a 4-point polygon."
            )

    @classmethod
    def from_extrema(cls, xmin: float, xmax: float, ymin: float, ymax: float):
        """
        Create a BoundingBox from extrema values.

        Parameters
        ----------
        xmin : float
            Minimum x-coordinate of the bounding box.
        xmax : float
            Maximum x-coordinate of the bounding box.
        ymin : float
            Minimum y-coordinate of the bounding box.
        ymax : float
            Maximum y-coordinate of the bounding box.

        Returns
        -------
        BoundingBox
            A BoundingBox created from the provided extrema values.
        """
        return cls(
            polygon=BasicPolygon(
                points=[
                    Point(x=xmin, y=ymin),
                    Point(x=xmax, y=ymin),
                    Point(x=xmax, y=ymax),
                    Point(x=xmin, y=ymax),
                ]
            )
        )

    @property
    def xmin(self):
        """Minimum x-coordinate of the bounding box."""
        return self.polygon.xmin

    @property
    def xmax(self):
        """Maximum x-coordinate of the bounding box."""
        return self.polygon.xmax

    @property
    def ymin(self):
        """Minimum y-coordinate of the bounding box."""
        return self.polygon.ymin

    @property
    def ymax(self):
        """Maximum y-coordinate of the bounding box."""
        return self.polygon.ymax


@dataclass
class MultiPolygon:
    """
    Represents a collection of polygons.

    Parameters
    ----------
    polygons : List[Polygon], optional
        List of `Polygon` objects. Defaults to an empty list.

    Raises
    ------
    TypeError
        If `polygons` is not a list, or an element in `polygons` is not a `Polygon`.

    Examples
    --------
    >>> MultiPolygon(
    ...     polygons=[
    ...         Polygon(...),
    ...         Polygon(...),
    ...         Polygon(...),
    ...     ]
    ... )
    """

    polygons: List[Polygon]

    def __post_init__(self):
        # unpack & validate
        if not isinstance(self.polygons, list):
            raise TypeError(
                "polygons should be list of `valor.schemas.Polyon`"
            )
        for i, polygon in enumerate(self.polygons):
            if isinstance(polygon, dict):
                self.polygons[i] = Polygon(**polygon)
            if not isinstance(self.polygons[i], Polygon):
                raise TypeError(
                    "polygons list should contain elements of type `valor.schemas.Polygon`"
                )


@dataclass
class Raster:
    """
    Represents a binary mask.

    Parameters
    ----------
    mask : str
        Base64-encoded string representing the raster mask.
    geometry : Union[BoundingBox, Polygon, MultiPolygon], optional
        Option to input the raster as a geometry. Overrides the mask.

    Raises
    ------
    TypeError
        If `encoding` is not a string.

    Examples
    --------
    Generate a random mask.
    >>> import numpy.random
    >>> height = 640
    >>> width = 480
    >>> array = numpy.random.rand(height, width)

    Convert to binary mask.
    >>> mask = (array > 0.5)

    Create Raster.
    >>> Raster.from_numpy(mask)
    """

    mask: str
    geometry: Union[BoundingBox, Polygon, MultiPolygon, None] = None

    def __post_init__(self):
        if not isinstance(self.mask, str):
            raise TypeError("mask should be of type `str`")
        if isinstance(self.geometry, dict):
            if "polygons" in self.geometry:
                self.geometry = MultiPolygon(**self.geometry)
            elif "boundary" in self.geometry:
                self.geometry = Polygon(**self.geometry)
            elif "polygon" in self.geometry:
                self.geometry = BoundingBox(**self.geometry)
            else:
                raise ValueError("Invalid dictionary structure.")

    @classmethod
    def from_numpy(cls, mask: np.ndarray):
        """
        Create a Raster object from a NumPy array.

        Parameters
        ----------
        mask : np.ndarray
            The 2D binary array representing the mask.

        Returns
        -------
        schemas.Raster

        Raises
        ------
        ValueError
            If the input array is not 2D or not of dtype bool.
        """
        if len(mask.shape) != 2:
            raise ValueError("raster currently only supports 2d arrays")
        if mask.dtype != bool:
            raise ValueError(
                f"Expecting a binary mask (i.e. of dtype bool) but got dtype {mask.dtype}"
            )
        f = io.BytesIO()
        PIL.Image.fromarray(mask).save(f, format="PNG")
        f.seek(0)
        mask_bytes = f.read()
        f.close()
        return cls(
            mask=b64encode(mask_bytes).decode(),
        )

    @classmethod
    def from_geometry(
        cls,
        geometry: Union[BoundingBox, Polygon, MultiPolygon],
        height: int,
        width: int,
    ):
        """
        Create a Raster object from a geometry.

        Parameters
        ----------
        geometry : Union[BoundingBox, Polygon, MultiPolygon]
            Defines the bitmask as a geometry. Overrides any existing mask.
        height : int
            The intended height of the binary mask.
        width : int
            The intended width of the binary mask.

        Returns
        -------
        schemas.Raster
        """
        r = cls.from_numpy(np.full((int(height), int(width)), False))
        r.geometry = geometry
        return r

    def to_numpy(self) -> np.ndarray:
        """
        Convert the base64-encoded mask to a NumPy array.

        Returns
        -------
        np.ndarray
            A 2D binary array representing the mask.
        """
        if self.geometry is not None:
            raise ValueError(
                "NumPy conversion is not supported for shape-based rasters."
            )
        mask_bytes = b64decode(self.mask)
        with io.BytesIO(mask_bytes) as f:
            img = PIL.Image.open(f)
            return np.array(img)



from .constraints import (
    BoolMapper,
    Constraint,
    DictionaryMapper,
    GeometryMapper,
    GeospatialMapper,
    LabelMapper,
    NumericMapper,
    StringMapper,
)
from .evaluation import EvaluationParameters, EvaluationRequest
from .filters import Filter
from .geometry import (
    BasicPolygon,
    BoundingBox,
    MultiPolygon,
    Point,
    Polygon,
    Raster,
)
from .metadata import dump_metadata, load_metadata, validate_metadata

__all__ = [
    "BasicPolygon",
    "Point",
    "BoundingBox",
    "Polygon",
    "MultiPolygon",
    "Raster",
    "EvaluationRequest",
    "EvaluationParameters",
    "Filter",
    "validate_metadata",
    "dump_metadata",
    "load_metadata",
    "Constraint",
    "BoolMapper",
    "NumericMapper",
    "StringMapper",
    "GeometryMapper",
    "GeospatialMapper",
    "DictionaryMapper",
    "LabelMapper",
]



import math
from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from PIL import Image, ImageDraw, ImageFont

from valor import Annotation, GroundTruth, Prediction, enums, schemas
from valor.metatypes import ImageMetadata

# https://sashat.me/2017/01/11/list-of-20-simple-distinct-colors/
COLOR_MAP = [
    (230, 25, 75),
    (60, 180, 75),
    (255, 225, 25),
    (0, 130, 200),
    (245, 130, 48),
    (145, 30, 180),
    (70, 240, 240),
    (240, 50, 230),
    (210, 245, 60),
    (250, 190, 190),
    (0, 128, 128),
    (230, 190, 255),
    (170, 110, 40),
    (255, 250, 200),
    (128, 0, 0),
    (170, 255, 195),
    (128, 128, 0),
    (255, 215, 180),
    (0, 0, 128),
    (128, 128, 128),
]


def _polygons_to_binary_mask(
    polys: List[schemas.Polygon], img_w, img_h
) -> np.ndarray:
    """note there's some aliasing/areas differences between this
    and e.g. postgis, so this method should only be used for visualization
    """
    mask = Image.new("1", (img_w, img_h), (False,))
    draw = ImageDraw.Draw(mask)
    for poly in polys:
        draw.polygon(poly.boundary.tuple_list(), fill=(True,))  # type: ignore
        if poly.holes is not None:
            for hole in poly.holes:
                draw.polygon(hole.tuple_list(), fill=(False,))  # type: ignore

    return np.array(mask)


def create_combined_segmentation_mask(
    annotated_datums: Sequence[Union[GroundTruth, Prediction]],
    label_key: str,
    task_type: Union[enums.TaskType, None] = None,
) -> Tuple[Image.Image, Dict[str, Image.Image]]:
    """
    Creates a combined segmentation mask from a list of segmentations.

    Parameters
    -------
    annotated_datums : List[Union[GroundTruth, Prediction]]
        A list of segmentations. These all must have the same `image` attribute.
    label_key : str
        The label key to use.
    task_type : enums.TaskType
        The associated task type.


    Returns
    -------
    tuple
        The first element of the tuple is the combined mask, as an RGB PIL image. The second
        element is a color legend: it's a dict with the unique labels as keys and the
        PIL image swatches as values.

    Raises
    ------
    RuntimeError
        If all segmentations don't belong to the same image or there is a
        segmentation that doesn't have `label_key` as the key of one of its labels.
    ValueError
        If there aren't any segmentations.
    """

    if len(annotated_datums) == 0:
        raise ValueError("`segs` cannot be empty.")

    if (
        len(
            set(
                [
                    annotated_datum.datum.uid
                    for annotated_datum in annotated_datums
                ]
            )
        )
        > 1
    ):
        raise RuntimeError(
            "Expected all segmentation to belong to the same image"
        )

    # Validate task type
    if task_type is not None and task_type not in [
        enums.TaskType.OBJECT_DETECTION,
        enums.TaskType.SEMANTIC_SEGMENTATION,
    ]:
        raise RuntimeError(
            "Expected either Instance or Semantic segmentation task_type."
        )

    # Create valid task type list
    if task_type is None:
        task_types = [
            enums.TaskType.OBJECT_DETECTION,
            enums.TaskType.SEMANTIC_SEGMENTATION,
        ]
    else:
        task_types = [task_type]

    # unpack raster annotations
    annotations: List[Annotation] = []
    for annotated_datum in annotated_datums:
        for annotation in annotated_datum.annotations:
            if annotation.task_type in task_types:
                annotations.append(annotation)

    label_values = []
    for annotation in annotations:
        for label in annotation.labels:
            if label.key == label_key:
                label_values.append(label.value)
    if not label_values:
        raise RuntimeError(
            f"Annotation doesn't have a label with key `{label_key}`"
        )

    unique_label_values = list(set(label_values))
    label_value_to_color = {
        v: COLOR_MAP[i] for i, v in enumerate(unique_label_values)
    }
    seg_colors = [label_value_to_color[v] for v in label_values]

    image = ImageMetadata.from_datum(annotated_datums[0].datum)
    img_w, img_h = image.width, image.height

    combined_mask = np.zeros((img_h, img_w, 3), dtype=np.uint8)
    for annotation, color in zip(annotations, seg_colors):
        if annotation.raster is not None:
            if annotation.raster.geometry is None:
                mask = annotation.raster.to_numpy()
            elif isinstance(annotation.raster.geometry, schemas.MultiPolygon):
                mask = _polygons_to_binary_mask(
                    annotation.raster.geometry.polygons,
                    img_w=img_w,
                    img_h=img_h,
                )
            elif isinstance(annotation.raster.geometry, schemas.Polygon):
                mask = _polygons_to_binary_mask(
                    [annotation.raster.geometry],
                    img_w=img_w,
                    img_h=img_h,
                )
            elif isinstance(annotation.raster.geometry, schemas.BoundingBox):
                mask = _polygons_to_binary_mask(
                    [
                        schemas.Polygon(
                            boundary=annotation.raster.geometry.polygon
                        )
                    ],
                    img_w=img_w,
                    img_h=img_h,
                )
            else:
                continue
        else:
            continue

        combined_mask[np.where(mask)] = color

    legend = {
        v: Image.new("RGB", (20, 20), color)
        for v, color in label_value_to_color.items()
    }

    return Image.fromarray(combined_mask), legend


def draw_detections_on_image(
    detections: Sequence[Union[GroundTruth, Prediction]],
    img: Image.Image,
) -> Image.Image:
    """
    Draws detections (bounding boxes and labels) on an image.

    Parameters
    -------
    detections : List[Union[GroundTruth, Prediction]]
        A list of `GroundTruths` or `Predictions` to draw on the image.
    img : Image.Image
        The image to draw the detections on.


    Returns
    -------
    img : Image.Image
        An image with the detections drawn on.
    """

    annotations = []
    for datum in detections:
        annotations.extend(datum.annotations)

    for i, detection in enumerate(annotations):
        if detection.task_type in [enums.TaskType.OBJECT_DETECTION]:
            img = _draw_detection_on_image(detection, img, inplace=i != 0)
    return img


def draw_bounding_box_on_image(
    bounding_box: schemas.BoundingBox,
    img: Image.Image,
    color: Tuple[int, int, int] = (255, 0, 0),
) -> Image.Image:
    """Draws a bounding polygon on an image. This operation is not done in place.

    Parameters
    ----------
    bounding_box
        Bounding box to draw on the image.
    img
        Pillow image to draw on.
    color
        RGB tuple of the color to use.

    Returns
    -------
    img
        Pillow image with bounding box drawn on it.
    """
    return _draw_bounding_polygon_on_image(
        bounding_box.polygon, img, color=color, inplace=False
    )


def _draw_detection_on_image(
    detection: Annotation, img: Image.Image, inplace: bool
) -> Image.Image:
    """Draw a detection on an image."""
    text = ", ".join(
        [f"{label.key}:{label.value}" for label in detection.labels]
    )
    if detection.polygon is not None:
        img = _draw_bounding_polygon_on_image(
            detection.polygon.boundary,
            img,
            inplace=inplace,
            text=text,
        )
    elif detection.bounding_box is not None:
        img = _draw_bounding_polygon_on_image(
            detection.bounding_box.polygon,
            img,
            inplace=inplace,
            text=text,
        )

    return img


def _draw_bounding_polygon_on_image(
    polygon: schemas.BasicPolygon,
    img: Image.Image,
    color: Tuple[int, int, int] = (255, 0, 0),
    inplace: bool = False,
    text: Optional[str] = None,
    font_size: int = 24,
) -> Image.Image:
    """Draw a bounding polygon on an image."""
    img = img if inplace else img.copy()
    img_draw = ImageDraw.Draw(img)

    img_draw.polygon(
        [(p.x, p.y) for p in polygon.points],
        outline=color,
    )

    if text is not None:
        _write_text(
            font_size=font_size,
            text=text,
            boundary=polygon,
            draw=img_draw,
            color=color,
        )
    return img


def _write_text(
    font_size: int,
    text: str,
    boundary: schemas.BasicPolygon,
    draw: ImageDraw.ImageDraw,
    color: Union[Tuple[int, int, int], str],
):
    """Write text on an image."""
    try:
        font = ImageFont.truetype("arial.ttf", font_size)
    except IOError:
        try:
            font = ImageFont.truetype("Arial.ttf", font_size)
        except IOError:
            font = ImageFont.load_default()

    _, _, text_width, text_height = font.getbbox(text)
    if boundary.ymin > text_height:
        text_bottom = boundary.ymin
    else:
        text_bottom = boundary.ymax + text_height

    margin = math.ceil(0.05 * text_height) - 1
    draw.rectangle(
        (
            (boundary.xmin, text_bottom - text_height - 2 * margin),
            (boundary.xmin + text_width, text_bottom),
        ),
        fill=color,
    )
    draw.text(
        (boundary.xmin + margin, text_bottom - text_height - margin),
        text,
        fill="black",
        font=font,
    )


def draw_raster_on_image(
    raster: schemas.Raster,
    img: Image.Image,
    color: Tuple[int, int, int] = (255, 0, 0),
    alpha: float = 0.4,
) -> Image.Image:
    """Draws the raster on top of an image. This operation is not done in place.

    Parameters
    ----------
    img
        pillow image to draw on.
    color
        RGB tuple of the color to use
    alpha
        alpha (transparency) value of the mask. 0 is fully transparent, 1 is fully opaque
    """
    img = img.copy()
    binary_mask = raster.to_numpy()
    mask_arr = np.zeros(
        (binary_mask.shape[0], binary_mask.shape[1], 3), dtype=np.uint8
    )
    mask_arr[binary_mask] = color
    mask_img = Image.fromarray(mask_arr)
    blend = Image.blend(img, mask_img, alpha=alpha)
    img.paste(blend, (0, 0), mask=Image.fromarray(binary_mask))

    return img



from typing import Optional, cast

import PIL.Image

from valor import Datum
from valor.schemas.metadata import MetadataType, validate_metadata


class ImageMetadata:
    """
    A class describing the metadata for a particular image.

    Parameters
    ----------
    uid : str
        The UID of the image.
    height : int
        The height of the image.
    width : int
        The width of the image.
    metadata : dict
        A dictionary of metadata that describes the image.
    """

    def __init__(
        self,
        uid: str,
        height: int,
        width: int,
        metadata: Optional[MetadataType] = None,
    ):
        self.uid = uid
        self._dataset_name = None
        self.height = height
        self.width = width
        self.metadata: MetadataType = dict(metadata) if metadata else {}

        if not isinstance(self.uid, str):
            raise TypeError("ImageMetadata uid must be a string.")
        if not isinstance(self.height, int):
            raise TypeError("ImageMetadata height must be a int.")
        if not isinstance(self.width, int):
            raise TypeError("ImageMetadata height must be a int.")
        validate_metadata(self.metadata)

    @staticmethod
    def valid(datum: Datum) -> bool:
        """
        Asserts whether the `Datum's` height and width is a valid subset of the image's height and width.

        Parameters
        ----------
        datum : Datum
            The `Datum` to check validity for.
        """
        return {"height", "width"}.issubset(datum.metadata)

    @classmethod
    def from_datum(cls, datum: Datum):
        """
        Creates an `ImageMetadata` object from a `Datum`.

        Parameters
        ----------
        datum : Datum
            The `Datum` to extract metadata from.
        """
        if not cls.valid(datum):
            raise ValueError(
                f"`datum` does not contain height and/or width in metadata `{datum.metadata}`"
            )
        metadata = dict(datum.metadata)
        width = cast(int, metadata.pop("width"))
        height = cast(int, metadata.pop("height"))
        img = cls(
            uid=datum.uid,
            height=int(height),
            width=int(width),
            metadata=metadata,
        )
        return img

    @classmethod
    def from_pil(cls, uid: str, image: PIL.Image.Image):
        """
        Creates an `ImageMetadata` object from an image.

        Parameters
        ----------
        uid : str
            The UID of the image.
        image : PIL.Image.Image
            The image to create metadata for.
        """
        width, height = image.size
        return cls(
            uid=uid,
            height=int(height),
            width=int(width),
        )

    def to_datum(self) -> Datum:
        """
        Converts an `ImageMetadata` object into a `Datum`.
        """
        metadata = dict(self.metadata) if self.metadata else {}

        metadata["height"] = self.height
        metadata["width"] = self.width
        datum = Datum(
            uid=self.uid,
            metadata=metadata,
        )
        return datum


class VideoFrameMetadata:
    """
    A class describing the metadata for the frame of a video.

    Parameters
    ----------
    image : ImageMetadata
        Metadata describing the frame of the video.
    frame: int
        The number of seconds into the video that the frame was taken.
    """

    def __init__(
        self,
        image: ImageMetadata,
        frame: int,
    ):
        self.image = image
        self.frame = frame

        if not isinstance(self.image, ImageMetadata):
            raise TypeError(
                "`image` should be of type `valor.metatypes.ImageMetadata`"
            )
        if not isinstance(self.frame, int):
            raise TypeError("`frame` should be of type `int`.")

    @staticmethod
    def valid(datum: Datum) -> bool:
        """
        Asserts whether the `Datum's` height and width is a valid subset of the image's height and width.

        Parameters
        ----------
        datum : Datum
            The `Datum` to check validity for.
        """
        return {"height", "width", "frame"}.issubset(datum.metadata)

    @classmethod
    def from_datum(cls, datum: Datum):
        """
        Creates a `VideoFrameMetadata` object from a `Datum`.

        Parameters
        ----------
        datum : Datum
            The `Datum` to extract metadata from.
        """
        if not cls.valid(datum):
            raise ValueError(
                f"`datum` does not contain height, width and/or frame in metadata `{datum.metadata}`"
            )
        image = ImageMetadata.from_datum(datum)
        frame = cast(int, image.metadata.pop("frame"))
        return cls(
            image=image,
            frame=int(frame),
        )

    def to_datum(self) -> Datum:
        """
        Converts a `VideoFrameMetadata` object into a `Datum`.
        """
        datum = self.image.to_datum()
        datum.metadata["frame"] = self.frame
        return datum



import datetime

import pytest

from valor.schemas.metadata import (
    dump_metadata,
    load_metadata,
    validate_metadata,
)


def test_validate_metadata():
    validate_metadata({"test": "test"})
    validate_metadata({"test": 1})
    validate_metadata({"test": 1.0})

    with pytest.raises(TypeError):
        validate_metadata({123: 123})  # type: ignore

    # Test supported value types
    with pytest.raises(TypeError):
        validate_metadata({"test": (1, 2)})  # type: ignore
    with pytest.raises(TypeError):
        validate_metadata({"test": [1, 2]})  # type: ignore

    # Check int to float conversion
    validate_metadata({"test": 1})


def test_dump_metadata():

    # should not be changed as only uses str, int, float values
    metadata = {
        "a": int(123.4),
        "b": float(123.4),
        "c": str(123.4),
    }
    assert metadata == dump_metadata(metadata)

    # check datetime object conversion
    metadata["d"] = datetime.datetime.fromisoformat("2023-01-01T12:12:12")
    metadata["e"] = datetime.date.fromisoformat("2023-01-01")
    metadata["f"] = datetime.time.fromisoformat("12:12:12:100000")
    metadata["g"] = datetime.timedelta(days=1)
    assert dump_metadata(metadata) == {
        "a": int(123.4),
        "b": float(123.4),
        "c": str(123.4),
        "d": {"datetime": "2023-01-01T12:12:12"},
        "e": {"date": "2023-01-01"},
        "f": {"time": "12:12:12.100000"},
        "g": {"duration": "86400.0"},
    }


def test_load_metadata():
    # should not be changed as only uses str, int, float values
    metadata = {
        "a": int(123.4),
        "b": float(123.4),
        "c": str(123.4),
    }
    assert metadata == load_metadata(metadata)

    # check datetime object conversion
    metadata["d"] = {"datetime": "2023-01-01T12:12:12"}
    metadata["e"] = {"date": "2023-01-01"}
    metadata["f"] = {"time": "12:12:12.100000"}
    metadata["g"] = {"duration": "86400.0"}
    assert load_metadata(metadata) == {
        "a": int(123.4),
        "b": float(123.4),
        "c": str(123.4),
        "d": datetime.datetime.fromisoformat("2023-01-01T12:12:12"),
        "e": datetime.date.fromisoformat("2023-01-01"),
        "f": datetime.time.fromisoformat("12:12:12:100000"),
        "g": datetime.timedelta(days=1),
    }



import datetime

import numpy
import pytest

from valor import Label
from valor.schemas.constraints import (
    DatetimeMapper,
    DictionaryMapper,
    GeometryMapper,
    GeospatialMapper,
    LabelMapper,
    NumericMapper,
    StringMapper,
    _DictionaryValueMapper,
)
from valor.schemas.geometry import (
    BasicPolygon,
    BoundingBox,
    MultiPolygon,
    Point,
    Polygon,
    Raster,
)


def _test_mapper(mapper, operator, value, retval=None, name_prefix=""):
    name = "name"
    key = "key"

    ret_name = f"{name_prefix}{name}"

    if operator == "==":
        expr = mapper(name, key) == value
    elif operator == "!=":
        expr = mapper(name, key) != value
    elif operator == ">=":
        expr = mapper(name, key) >= value
    elif operator == "<=":
        expr = mapper(name, key) <= value
    elif operator == ">":
        expr = mapper(name, key) > value
    elif operator == "<":
        expr = mapper(name, key) < value
    elif operator == "contains":
        expr = mapper(name, key).contains(value)
    elif operator == "intersect":
        expr = mapper(name, key).intersect(value)
    elif operator == "inside":
        expr = mapper(name, key).inside(value)
    elif operator == "outside":
        expr = mapper(name, key).outside(value)
    elif operator == "is_none":
        expr = mapper(name, key).is_none()
        value = None
    elif operator == "exists":
        expr = mapper(name, key).exists()
        value = None
    elif operator == "in_":
        expr = mapper(name, key).in_([value])[0]
        operator = "=="
    else:
        raise NotImplementedError

    assert expr.name == ret_name
    assert expr.key == key
    assert expr.constraint.operator == operator

    if retval:
        assert expr.constraint.value == retval
    elif value is None:
        assert expr.constraint.value is None
    else:
        assert expr.constraint.value == value


def test_numeric_mapper():

    # test operators on `int` and `float`
    for value in [int(123), float(0.123)]:

        def _test_numeric_mapper(operator):
            _test_mapper(NumericMapper, operator, value)

        _test_numeric_mapper("==")
        _test_numeric_mapper("!=")
        _test_numeric_mapper(">=")
        _test_numeric_mapper("<=")
        _test_numeric_mapper(">")
        _test_numeric_mapper("<")
        _test_numeric_mapper("in_")

        with pytest.raises(AttributeError):
            _test_numeric_mapper("is_none")
        with pytest.raises(AttributeError):
            _test_numeric_mapper("exists")
        with pytest.raises(AttributeError):
            _test_numeric_mapper("contains")
        with pytest.raises(AttributeError):
            _test_numeric_mapper("intersect")
        with pytest.raises(AttributeError):
            _test_numeric_mapper("inside")
        with pytest.raises(AttributeError):
            _test_numeric_mapper("outside")

    # test invalid type
    with pytest.raises(TypeError):
        _test_mapper(NumericMapper, "==", "some_str")


def test_string_mapper():

    # test operators on `str`
    def _test_string_mapper(operator):
        _test_mapper(StringMapper, operator, "some_str")

    _test_string_mapper("==")
    _test_string_mapper("!=")
    _test_string_mapper("in_")

    with pytest.raises(AttributeError):
        _test_string_mapper(">=")
    with pytest.raises(AttributeError):
        _test_string_mapper("<=")
    with pytest.raises(AttributeError):
        _test_string_mapper(">")
    with pytest.raises(AttributeError):
        _test_string_mapper("<")
    with pytest.raises(AttributeError):
        _test_string_mapper("is_none")
    with pytest.raises(AttributeError):
        _test_string_mapper("exists")
    with pytest.raises(AttributeError):
        _test_string_mapper("contains")
    with pytest.raises(AttributeError):
        _test_string_mapper("intersect")
    with pytest.raises(AttributeError):
        _test_string_mapper("inside")
    with pytest.raises(AttributeError):
        _test_string_mapper("outside")

    # test invalid type
    with pytest.raises(TypeError):
        _test_mapper(StringMapper, "==", 123)


def test_datetime_mapper():

    # test operators on `datetime` objects
    datetime_ = datetime.datetime.now()
    date_ = datetime.date.today()
    time_ = datetime.datetime.now().time()
    timedelta_ = datetime.timedelta(days=1)

    datetime_filter = {"datetime": str(datetime_.isoformat())}
    date_filter = {"date": str(date_.isoformat())}
    time_filter = {"time": str(time_.isoformat())}
    timedelta_filter = {"duration": str(timedelta_.total_seconds())}

    for value, retval in [
        (datetime_, datetime_filter),
        (date_, date_filter),
        (time_, time_filter),
        (timedelta_, timedelta_filter),
    ]:

        def _test_datetime_mapper(operator):
            _test_mapper(DatetimeMapper, operator, value, retval=retval)

        _test_datetime_mapper("==")
        _test_datetime_mapper("!=")
        _test_datetime_mapper(">=")
        _test_datetime_mapper("<=")
        _test_datetime_mapper(">")
        _test_datetime_mapper("<")
        _test_datetime_mapper("in_")

        with pytest.raises(AttributeError):
            _test_datetime_mapper("is_none")
        with pytest.raises(AttributeError):
            _test_datetime_mapper("exists")
        with pytest.raises(AttributeError):
            _test_datetime_mapper("contains")
        with pytest.raises(AttributeError):
            _test_datetime_mapper("intersect")
        with pytest.raises(AttributeError):
            _test_datetime_mapper("inside")
        with pytest.raises(AttributeError):
            _test_datetime_mapper("outside")

    # test invalid type
    with pytest.raises(TypeError):
        _test_mapper(DatetimeMapper, "==", 1234)


def test_geometry_mapper():

    # test attribute mappers
    assert type(GeometryMapper("name", "key").area) is NumericMapper

    # test operators on geometry objects
    point = Point(1, 1)
    basic_polygon = BasicPolygon(
        points=[
            Point(0, 0),
            Point(0, 1),
            Point(1, 0),
        ]
    )
    polygon = Polygon(
        boundary=basic_polygon,
    )
    bbox = BoundingBox.from_extrema(xmin=0, xmax=1, ymin=0, ymax=1)
    multipolygon = MultiPolygon(polygons=[polygon])
    raster = Raster.from_numpy(numpy.zeros((10, 10)) == True)  # noqa 712

    for value in [
        point,
        bbox,
        polygon,
        multipolygon,
        raster,
    ]:

        def _test_spatial_mapper(operator, name_prefix=""):
            _test_mapper(
                GeometryMapper, operator, value, name_prefix=name_prefix
            )

        _test_spatial_mapper("is_none", "require_")
        _test_spatial_mapper("exists", "require_")

        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("contains")
        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("intersect")
        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("inside")
        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("outside")

        with pytest.raises(AttributeError):
            _test_spatial_mapper("==")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("!=")
        with pytest.raises(AttributeError):
            _test_spatial_mapper(">=")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("<=")
        with pytest.raises(AttributeError):
            _test_spatial_mapper(">")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("<")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("in_")

    # test unsupported geometry
    with pytest.raises(TypeError):
        _test_mapper(GeometryMapper, "intersect", "intersect")

    # test invalid type
    with pytest.raises(TypeError):
        _test_mapper(GeometryMapper, "intersect", "some_str")


def test_geospatial_mapper():

    # test attribute mappers
    assert type(GeospatialMapper("name", "key").area) is NumericMapper

    # test operators on geojson
    point = {"type": "Point", "coordinates": [30.0, 10.0]}

    for value in [
        point,
    ]:

        def _test_spatial_mapper(operator):
            _test_mapper(GeospatialMapper, operator, value)

        _test_spatial_mapper("intersect")
        _test_spatial_mapper("inside")
        _test_spatial_mapper("outside")

        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("contains")
        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("is_none")
        with pytest.raises(NotImplementedError):
            _test_spatial_mapper("exists")

        with pytest.raises(AttributeError):
            _test_spatial_mapper("==")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("!=")
        with pytest.raises(AttributeError):
            _test_spatial_mapper(">=")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("<=")
        with pytest.raises(AttributeError):
            _test_spatial_mapper(">")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("<")
        with pytest.raises(AttributeError):
            _test_spatial_mapper("in_")

    # test invalid type
    with pytest.raises(TypeError):
        _test_mapper(GeospatialMapper, "intersect", "some_str")


def test_dictionary_mapper():
    def mapper(operator, key) -> _DictionaryValueMapper:
        return DictionaryMapper("name")[key]

    # test operators on `int` and `float`
    for value in [int(123), float(0.123)]:

        def _test_dict_mapper(operator):
            _test_mapper(mapper, operator, value)

        _test_dict_mapper("is_none")
        _test_dict_mapper("exists")
        _test_dict_mapper("==")
        _test_dict_mapper("!=")
        _test_dict_mapper(">=")
        _test_dict_mapper("<=")
        _test_dict_mapper(">")
        _test_dict_mapper("<")
        _test_dict_mapper("in_")

        with pytest.raises(AttributeError):
            _test_dict_mapper("contains")
        with pytest.raises(AttributeError):
            _test_dict_mapper("intersect")
        with pytest.raises(AttributeError):
            _test_dict_mapper("inside")
        with pytest.raises(AttributeError):
            _test_dict_mapper("outside")

    # test operators on `str`
    value = "some_str"

    def _test_dict_mapper(operator):
        _test_mapper(mapper, operator, value)

    _test_dict_mapper("is_none")
    _test_dict_mapper("exists")
    _test_dict_mapper("==")
    _test_dict_mapper("!=")
    _test_dict_mapper("in_")

    with pytest.raises(AttributeError):
        _test_dict_mapper(">=")
    with pytest.raises(AttributeError):
        _test_dict_mapper("<=")
    with pytest.raises(AttributeError):
        _test_dict_mapper(">")
    with pytest.raises(AttributeError):
        _test_dict_mapper("<")
    with pytest.raises(AttributeError):
        _test_dict_mapper("contains")
    with pytest.raises(AttributeError):
        _test_dict_mapper("intersect")
    with pytest.raises(AttributeError):
        _test_dict_mapper("inside")
    with pytest.raises(AttributeError):
        _test_dict_mapper("outside")

    # test operators on `datetime` objects
    datetime_ = datetime.datetime.now()
    date_ = datetime.date.today()
    time_ = datetime.datetime.now().time()
    timedelta_ = datetime.timedelta(days=1)

    datetime_filter = {"datetime": str(datetime_.isoformat())}
    date_filter = {"date": str(date_.isoformat())}
    time_filter = {"time": str(time_.isoformat())}
    timedelta_filter = {"duration": str(timedelta_.total_seconds())}

    for value, retval in [
        (datetime_, datetime_filter),
        (date_, date_filter),
        (time_, time_filter),
        (timedelta_, timedelta_filter),
    ]:

        def _test_dict_mapper_no_key(operator):
            _test_mapper(mapper, operator, value)

        _test_dict_mapper_no_key("is_none")
        _test_dict_mapper_no_key("exists")

        def _test_dict_mapper(operator):
            _test_mapper(mapper, operator, value, retval=retval)

        _test_dict_mapper("==")
        _test_dict_mapper("!=")
        _test_dict_mapper(">=")
        _test_dict_mapper("<=")
        _test_dict_mapper(">")
        _test_dict_mapper("<")
        _test_dict_mapper("in_")

        with pytest.raises(AttributeError):
            _test_dict_mapper("contains")
        with pytest.raises(AttributeError):
            _test_dict_mapper("intersect")
        with pytest.raises(AttributeError):
            _test_dict_mapper("inside")
        with pytest.raises(AttributeError):
            _test_dict_mapper("outside")

    # test invalid type
    with pytest.raises(NotImplementedError):
        DictionaryMapper("name")["key"] == Point(1, 1)  # type: ignore


def test_label_mapper():

    # test operators on `Label`
    label = Label(key="k1", value="v1")

    def _test_label_mapper(operator):
        _test_mapper(LabelMapper, operator, label, retval={"k1": "v1"})

    _test_label_mapper("==")
    _test_label_mapper("!=")
    _test_label_mapper("in_")

    with pytest.raises(AttributeError):
        _test_label_mapper(">=")
    with pytest.raises(AttributeError):
        _test_label_mapper("<=")
    with pytest.raises(AttributeError):
        _test_label_mapper(">")
    with pytest.raises(AttributeError):
        _test_label_mapper("<")
    with pytest.raises(AttributeError):
        _test_label_mapper("is_none")
    with pytest.raises(AttributeError):
        _test_label_mapper("exists")
    with pytest.raises(AttributeError):
        _test_label_mapper("contains")
    with pytest.raises(AttributeError):
        _test_label_mapper("intersect")
    with pytest.raises(AttributeError):
        _test_label_mapper("inside")
    with pytest.raises(AttributeError):
        _test_label_mapper("outside")

    # test invalid type
    with pytest.raises(TypeError):
        _test_mapper(LabelMapper, "==", 123)



import datetime
from typing import Dict, List, Union

from valor.typing import is_geojson

GeoJSONPointType = Dict[str, Union[str, List[Union[float, int]]]]
GeoJSONPolygonType = Dict[str, Union[str, List[List[List[Union[float, int]]]]]]
GeoJSONMultiPolygonType = Dict[
    str, Union[str, List[List[List[List[Union[float, int]]]]]]
]
GeoJSONType = Union[
    GeoJSONPointType, GeoJSONPolygonType, GeoJSONMultiPolygonType
]

ValueType = Union[
    bool,
    int,
    float,
    str,
    datetime.datetime,
    datetime.date,
    datetime.time,
    datetime.timedelta,
    GeoJSONType,
]

MetadataType = Dict[str, ValueType]
ConvertibleMetadataType = Dict[
    str,
    Union[
        bool,
        int,
        float,
        str,
        Dict[str, str],
        Dict[str, GeoJSONType],
    ],
]


def _convert_object_to_metadatum(
    value: ValueType,
) -> Union[bool, int, float, str, Dict[str, str], Dict[str, GeoJSONType]]:
    """Converts an object into a valor metadatum."""

    # atomic types
    if (
        isinstance(value, bool)
        or isinstance(value, int)
        or isinstance(value, float)
        or isinstance(value, str)
    ):
        return value

    # datetime
    elif isinstance(value, datetime.datetime):
        return {"datetime": value.isoformat()}
    elif isinstance(value, datetime.date):
        return {"date": value.isoformat()}
    elif isinstance(value, datetime.time):
        return {"time": value.isoformat()}
    elif isinstance(value, datetime.timedelta):
        return {"duration": str(value.total_seconds())}

    # geojson
    elif is_geojson(value):
        return {"geojson": value}

    # not implemented
    else:
        raise NotImplementedError(
            f"Object with type '{type(value)}' is not currently supported."
        )


def _convert_metadatum_to_object(
    value: Union[bool, int, float, str, Dict[str, str], Dict[str, GeoJSONType]]
) -> ValueType:
    """Converts a valor metadatum into an object."""

    # atomic types
    if (
        isinstance(value, bool)
        or isinstance(value, int)
        or isinstance(value, float)
        or isinstance(value, str)
    ):
        return value

    # validate serialized type
    elif not isinstance(value, dict):
        raise TypeError(
            f"Object with type '{type(value)}' is not atomic and not serialized."
        )

    # datetime
    elif "datetime" in value:
        if not isinstance(value["datetime"], str):
            raise TypeError
        return datetime.datetime.fromisoformat(value["datetime"])
    elif "date" in value:
        if not isinstance(value["date"], str):
            raise TypeError
        return datetime.date.fromisoformat(value["date"])
    elif "time" in value:
        if not isinstance(value["time"], str):
            raise TypeError
        return datetime.time.fromisoformat(value["time"])
    elif "duration" in value:
        if not isinstance(value["duration"], str):
            raise TypeError
        return datetime.timedelta(seconds=float(value["duration"]))

    # geojson
    elif "geojson" in value:
        return value["geojson"]

    # not implemented
    else:
        raise TypeError(
            f"Object with type '{value.keys()}' is not currently supported."
        )


def validate_metadata(metadata: MetadataType):
    """Validates metadata dictionary."""
    if not isinstance(metadata, dict):
        raise TypeError("`metadata` should be an object of type `dict`.")
    for key, value in metadata.items():
        if not isinstance(key, str):
            raise TypeError("`metadata` key should have type `str`.")
        if not (
            isinstance(value, int)
            or isinstance(value, float)
            or isinstance(value, str)
            or isinstance(value, datetime.datetime)
            or isinstance(value, datetime.date)
            or isinstance(value, datetime.time)
            or is_geojson(value)
        ):
            raise TypeError(
                "`metadata` value should have type `str`, `int`, `float`, `datetime` or `geojson`."
            )


def dump_metadata(metadata: MetadataType) -> ConvertibleMetadataType:
    """Converts metadata to API-compatible dictionary."""
    return {
        key: _convert_object_to_metadatum(value)
        for key, value in metadata.items()
    }


def load_metadata(metadata: ConvertibleMetadataType) -> MetadataType:
    """Converts API metadata to Client-compatible dictionary."""
    return {
        key: _convert_metadatum_to_object(value)
        for key, value in metadata.items()
    }



import numpy as np
import pytest

from valor import schemas


def test_point():
    # valid
    p1 = schemas.Point(x=1, y=1)
    p2 = schemas.Point(x=1.0, y=1.0)
    p3 = schemas.Point(x=1.0, y=0.99)

    # test member fn `__hash__`
    assert p1.__hash__() == p2.__hash__()
    assert p1.__hash__() != p3.__hash__()

    # test member fn `resize`
    p11 = p1.resize(
        og_img_h=10,
        og_img_w=10,
        new_img_h=100,
        new_img_w=100,
    )
    assert p11.x == p1.x * 10
    assert p11.y == p1.y * 10


def test_basicpolygon():
    p1 = schemas.Point(x=0, y=0)
    p2 = schemas.Point(x=5, y=0)
    p3 = schemas.Point(x=10, y=5)
    p4 = schemas.Point(x=0, y=5)

    schemas.BasicPolygon(points=[p1, p2, p4])

    # Test __post_init__
    with pytest.raises(TypeError) as e:
        schemas.BasicPolygon(points=p1)  # type: ignore
    assert "is not a list" in str(e)
    with pytest.raises(TypeError) as e:
        schemas.BasicPolygon(points=[p1, p2, 4])  # type: ignore
    assert "is not a `Point`"
    with pytest.raises(ValueError) as e:
        schemas.BasicPolygon(points=[p1, p2])
    assert "needs at least 3 unique points" in str(e)

    # Test member fn `xy_list`
    poly = schemas.BasicPolygon(points=[p1, p2, p3])
    assert poly.xy_list() == [p1, p2, p3]

    # Test properties
    assert poly.xmin == 0
    assert poly.xmax == 10
    assert poly.ymin == 0
    assert poly.ymax == 5


def test_polygon():
    p1 = schemas.Point(x=0, y=0)
    p2 = schemas.Point(x=5, y=0)
    p3 = schemas.Point(x=0, y=5)
    poly = schemas.BasicPolygon(points=[p1, p2, p3])

    # valid
    schemas.Polygon(boundary=poly)
    schemas.Polygon(boundary=poly, holes=[poly])

    # test `__post_init__`
    with pytest.raises(TypeError) as e:
        schemas.Polygon(boundary=123)  # type: ignore
    assert "boundary should be of type `valor.schemas.BasicPolygon`" in str(e)
    with pytest.raises(TypeError) as e:
        schemas.Polygon(boundary=poly, holes=123)  # type: ignore
    assert "holes should be a list of `valor.schemas.BasicPolygon`" in str(e)
    with pytest.raises(TypeError) as e:
        schemas.Polygon(boundary=poly, holes=[123])  # type: ignore
    assert (
        "should contain elements of type `valor.schemas.BasicPolygon`"
        in str(e)
    )


def test_boundingbox():
    p1 = schemas.Point(x=0, y=0)
    p2 = schemas.Point(x=5, y=0)
    p3 = schemas.Point(x=5, y=5)
    p4 = schemas.Point(x=0, y=5)
    poly = schemas.BasicPolygon(points=[p1, p2, p3, p4])

    # test __post_init__
    schemas.BoundingBox(polygon=poly)
    with pytest.raises(TypeError) as e:
        schemas.BoundingBox(polygon=p1)  # type: ignore
    assert "should be of type `valor.schemas.BasicPolygon`" in str(e)
    with pytest.raises(ValueError) as e:
        schemas.BoundingBox(polygon=schemas.BasicPolygon(points=[p1, p2, p3]))
    assert "should be made of a 4-point polygon" in str(e)

    # test classmethod `from_extrema`
    bbox = schemas.BoundingBox.from_extrema(xmin=-1, xmax=10, ymin=-2, ymax=11)
    assert bbox.polygon.xy_list() == [
        schemas.Point(x=-1, y=-2),
        schemas.Point(x=10, y=-2),
        schemas.Point(x=10, y=11),
        schemas.Point(x=-1, y=11),
    ]


def test_multipolygon():
    p1 = schemas.Point(x=0, y=0)
    p2 = schemas.Point(x=5, y=0)
    p3 = schemas.Point(x=5, y=5)
    p4 = schemas.Point(x=0, y=5)
    component_poly = schemas.BasicPolygon(points=[p1, p2, p3, p4])
    poly = schemas.Polygon(boundary=component_poly)

    # valid
    schemas.MultiPolygon(polygons=[poly])

    # test `__post_init__`
    with pytest.raises(TypeError) as e:
        schemas.MultiPolygon(polygons=component_poly)  # type: ignore
    assert "polygons should be list of `valor.schemas.Polyon`" in str(e)
    with pytest.raises(TypeError) as e:
        schemas.MultiPolygon(polygons=[component_poly])  # type: ignore
    assert (
        "polygons list should contain elements of type `valor.schemas.Polygon`"
        in str(e)
    )


def test_raster(raster_raw_mask):
    mask1 = np.ones((10, 10)) == 1

    # valid
    schemas.Raster(mask="test")
    schemas.Raster.from_numpy(mask=mask1)

    # test `__post_init__`
    with pytest.raises(TypeError) as e:
        schemas.Raster(mask=123)  # type: ignore
    assert "mask should be of type `str`" in str(e)

    # test classmethod `from_numpy`
    mask2 = np.ones((10, 10, 10)) == 1
    mask3 = np.ones((10, 10))
    with pytest.raises(ValueError) as e:
        schemas.Raster.from_numpy(mask2)
    assert "raster currently only supports 2d arrays" in str(e)
    with pytest.raises(ValueError) as e:
        schemas.Raster.from_numpy(mask3)
    assert "Expecting a binary mask" in str(e)

    # test member fn `to_numpy`
    r = schemas.Raster.from_numpy(raster_raw_mask)
    assert (
        r.mask
        == "iVBORw0KGgoAAAANSUhEUgAAABQAAAAUAQAAAACl8iCgAAAAF0lEQVR4nGP4f4CBiYGBIGZgsP9AjDoAuysDE0GVDN8AAAAASUVORK5CYII="
    )
    assert (r.to_numpy() == raster_raw_mask).all()
